/** ********************************************************************************************************************
 * @file        HTTP.ino
 * @author      Briowireless, Inc.
 * @date        2017-01-09
 * @brief       Main program file for a sample that uses all the BitPipe API.
 * @version     API V1
 * @copyright   Briowireless, Inc., 2017, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/

// Include directives  -------------------------------------------------------------------------------------------------

#include <bp_api.h>

// Preprocessor definitions --------------------------------------------------------------------------------------------

#define LED_PIN                             13

#define MAX_PAYLOAD_SIZE                    260

#ifdef ARDUINO_SAM_ZERO
#define SERIAL_PORT                         Serial1             // Serial class instance on which the BitPipe is found.
#else
#define SERIAL_PORT                         Serial              // Serial class instance on which the BitPipe is found.
#endif
#if (F_CPU > 8000000L)
#define SERIAL_SPEED                        115200
#else
#define SERIAL_SPEED                        9600
#endif
#define SERIAL_CONFIG                       SERIAL_8N1

#define REQUEST_TIMEOUT                     (30*1000UL)

#define SIM_PIN                             ""                  // When SIM card is locked using a PIN
#define SIM_PUK                             ""                  // When too many bad PINs were tried, the PIN Unlock Key is necessary.

#define APN_CONFIG_HOST                     ""                  // TODO: Enter you APN host address here
#define APN_CONFIG_USER                     ""
#define APN_CONFIG_PASS                     ""

#define HTTP_REMOTE_HOST                    "httpbin.org"
#define HTTP_REMOTE_PORT                    80
#define HTTP_PATH_QUERY                     "/post"
#define HTTP_ADD_HEADERS                    "MyCustomHeader: A simple value\\0d\\0aMyOtherHeader: AnotherValue"

#define HTTP_TMP_BLOCK_LEN                  32
#define HTTP_POST_DATA_LEN                  684

#define HTTP_CONN_STATUS_CLOSED             0
#define HTTP_CONN_STATUS_OPEN               1
#define HTTP_CONN_STATUS_ESTABLISHED        2

typedef enum 
{
    POWER_UP,
    SET_MAX_PAYLOAD_SIZE,
    GET_HW_VERSION,
    GET_FW_VERSION,
    POWER_ON_RADIO,
    GET_RADIO_STATUS,
    WAIT_RADIO_READY,
    SET_SIM_ACCESS,
    SET_APN_CONFIG,
    GET_MOBILE_CONN_STATUS,
    WAIT_NETWORK_REGISTRATION,
    HTTP_CONN_INIT,
    HTTP_WAIT_CONN,
    HTTP_POST_WRITE,
    HTTP_POST_SEND,
    HTTP_WAIT_DATA,
    DONE,
    NO_SIM_DETECTED,
}State;

// Local functions prototypes ------------------------------------------------------------------------------------------

// Callbacks
void sys_set_max_payload_sz_reply(uint8_t seq_nr, uint8_t status);
void devi_get_hw_version_reply(uint8_t seq_nr, uint8_t status, uint8_t hw_ver_len, uint8_t* hw_ver);
void devi_get_fw_version_reply(uint8_t seq_nr, uint8_t status, uint16_t boot_ver, uint8_t api_ver,
                               uint8_t bp_ver_len, uint8_t* bp_ver, uint8_t radio_ver_len, uint8_t* radio_ver);
void mobi_power_on_radio_reply(uint8_t seq_nr, uint8_t status);
void mobi_radio_start_notify_post(uint8_t seq_nr, uint8_t notify_reason, uint8_t sim_sta);
void mobi_get_radio_status_reply(uint8_t seq_nr, uint8_t status, uint8_t radio_status, uint8_t sim_sta);
void mobi_set_sim_access_config_reply(uint8_t seq_nr, uint8_t status);
void mobi_set_apn_config_reply(uint8_t seq_nr, uint8_t status);
void mobi_get_mobile_conn_status_reply(uint8_t seq_nr, uint8_t status, uint8_t conn_ind, uint8_t link_tech,
                                              uint16_t net_lac, uint32_t net_cell_id, uint8_t signal_strength,
                                              uint8_t* net_mcc, uint8_t* net_mnc);
void mobi_mobile_conn_status_change_notify_post(uint8_t seq_nr, uint8_t conn_ind, uint8_t link_tech,
                                                       uint16_t net_lac, uint32_t net_cell_id);
void ip_http_init_connection_reply(uint8_t seq_nr, uint8_t status, uint8_t conn_id);
void ip_http_post_write_ready_notify_post(uint8_t seq_nr, uint8_t conn_id);
void ip_http_post_write_reply(uint8_t seq_nr, uint8_t status);
void ip_http_rx_notify_request(uint8_t seq_nr, uint8_t conn_id, uint8_t data_len, uint8_t* data);
void ip_connection_close_notify_post(uint8_t seq_nr, uint8_t conn_id, uint8_t reason);

void uart_write(const uint8_t* data, uint16_t data_len);

// Local variables  ----------------------------------------------------------------------------------------------------

// Buffers used by SDK.
uint8_t deser_buf[MAX_PAYLOAD_SIZE + MESSAGE_PROTOCOL_HEADER_SIZE + MESSAGE_PROTOCOL_FOOTER_SIZE];
uint8_t ser_buf[MAX_PAYLOAD_SIZE];

bp_api_init_params init_params;

State state;
mp_seq_nr seq_number;
bool awaiting_reply;
unsigned long request_ts;
bool radio_ready;
uint8_t sim_status;
bool network_registered;

uint8_t http_conn_id;
uint8_t http_conn_status;

char http_tmp_buffer[HTTP_TMP_BLOCK_LEN];
uint8_t http_bytes_written;
size_t http_bytes_sent;

#ifdef ARDUINO_ARCH_AVR
const char http_post_data[HTTP_POST_DATA_LEN+1] PROGMEM =
#else
const char http_post_data[HTTP_POST_DATA_LEN+1] =
#endif
{
    "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Proin sagittis risus eu est aliquam, "
    "a pretium sapien pharetra. Quisque magna dolor, semper sit amet bibendum ut, laoreet id nisi. "
    "Morbi eleifend eros risus. Donec vitae risus felis. Etiam fermentum nisl sit amet vestibulum "
    "scelerisque. Donec condimentum justo nec enim malesuada, at varius ipsum viverra. Aliquam sapien "
    "elit, convallis id molestie sit amet, ornare non sem. Nam lacinia nulla nec ligula maximus gravida. "
    "Ut in odio ut turpis lacinia consectetur a ac quam. Ut volutpat pharetra massa sit amet rhoncus. "
    "Sed nec ante non ligula ullamcorper fringilla. Suspendisse malesuada metus vel ligula sollicitudin facilisis."
};

/* Unused callbacks can be set to NULL */
bp_api_callbacks callbacks  = {
    NULL,                                       // bp_api_sys_error_report_post
    sys_set_max_payload_sz_reply,               // bp_api_sys_set_max_payload_sz_reply
    NULL,                                       // bp_api_sys_set_uart_speed_reply

    devi_get_hw_version_reply,                  // bp_api_devi_get_hw_version_reply
    devi_get_fw_version_reply,                  // bp_api_devi_get_fw_version_reply
    NULL,                                       // bp_api_devi_get_product_name_reply
    NULL,                                       // bp_api_devi_get_vin_vio_voltage_reply
    NULL,                                       // bp_api_devi_get_junction_temp_reply
    NULL,                                       // bp_api_devi_set_low_power_reply
    NULL,                                       // bp_api_devi_get_latest_fw_version_reply
    NULL,                                       // bp_api_devi_update_fw_reply
    NULL,                                       // bp_api_devi_get_autonomous_mode_reply
    NULL,                                       // bp_api_devi_set_autonomous_mode_reply

    mobi_power_on_radio_reply,                  // bp_api_mobi_power_on_radio_reply
    NULL,                                       // bp_api_mobi_power_off_radio_reply
    mobi_radio_start_notify_post,               // bp_api_mobi_radio_start_notify_post
    mobi_get_radio_status_reply,                // bp_api_mobi_get_radio_status_reply
    NULL,                                       // bp_api_mobi_get_imei_reply
    NULL,                                       // bp_api_mobi_get_iccid_reply
    NULL,                                       // bp_api_mobi_get_imsi_reply
    NULL,                                       // bp_api_mobi_get_own_phone_number_reply
    NULL,                                       // bp_api_mobi_get_apn_config_reply
    mobi_set_apn_config_reply,                  // bp_api_mobi_set_apn_config_reply
    mobi_set_sim_access_config_reply,           // bp_api_mobi_set_sim_access_config_reply
    mobi_get_mobile_conn_status_reply,          // bp_api_mobi_get_mobile_conn_status_reply
    mobi_mobile_conn_status_change_notify_post, // bp_api_mobi_mobile_conn_status_change_notify_post
    NULL,                                       // bp_api_mobi_get_mobile_date_time_reply

    NULL,                                       // bp_api_ip_tcp_init_connection_reply
    NULL,                                       // bp_api_ip_tcp_connection_established_notify_post
    NULL,                                       // bp_api_ip_tcp_write_append_reply
    NULL,                                       // bp_api_ip_tcp_write_send_reply
    NULL,                                       // bp_api_ip_tcp_rx_notify_request
    ip_http_init_connection_reply,              // bp_api_ip_http_init_connection_reply
    ip_http_post_write_ready_notify_post,       // bp_api_ip_http_post_write_ready_notify_post
    ip_http_post_write_reply,                   // bp_api_ip_http_post_write_reply
    ip_http_rx_notify_request,                  // bp_api_ip_http_rx_notify_request
    NULL,                                       // bp_api_ip_connection_close_reply
    ip_connection_close_notify_post,            // bp_api_ip_connection_close_notify_post

    NULL,                                       // bp_api_mqtt_init_connection_reply
    NULL,                                       // bp_api_mqtt_connection_established_notify_post
    NULL,                                       // bp_api_mqtt_subscribe_topics_reply
    NULL,                                       // bp_api_mqtt_unsubscribe_topics_reply
    NULL,                                       // bp_api_mqtt_publish_reply
    NULL,                                       // bp_api_mqtt_publication_rx_notify_request
    NULL,                                       // bp_api_mqtt_disconnect_reply
    NULL,                                       // bp_api_mqtt_disconnect_notify_post

    NULL,                                       // bp_api_sms_setup_transmission_reply
    NULL,                                       // bp_api_sms_message_append_send_reply
    NULL,                                       // bp_api_sms_message_rx_notify_request
    NULL,                                       // bp_api_sms_message_part_notify_request
};

// Local functions  ----------------------------------------------------------------------------------------------------

void sys_set_max_payload_sz_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = GET_HW_VERSION;
        awaiting_reply = false;
    }
}

void devi_get_hw_version_reply(uint8_t seq_nr, uint8_t status, uint8_t hw_ver_len, uint8_t* hw_ver)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = GET_FW_VERSION;
        awaiting_reply = false;
    }
}

void devi_get_fw_version_reply(uint8_t seq_nr, uint8_t status, uint16_t boot_ver, uint8_t api_ver,
                               uint8_t bp_ver_len, uint8_t* bp_ver, uint8_t radio_ver_len, uint8_t* radio_ver)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = POWER_ON_RADIO;
        awaiting_reply = false;
    }
}

void mobi_power_on_radio_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        (status == MSG_STATUS_CODE_SUCCESS || status == BP_API_PWRON_RADIO__STATUS_ALREADY_STARTED))
    {
        if (status == MSG_STATUS_CODE_SUCCESS)
        {
            radio_ready = false;
            state = GET_RADIO_STATUS;
        }
        else if (status == BP_API_PWRON_RADIO__STATUS_ALREADY_STARTED)
        {
            radio_ready = true;
            state = SET_SIM_ACCESS;
        }
        awaiting_reply = false;
    }
}

void mobi_radio_start_notify_post(uint8_t seq_nr, uint8_t notify_reason, uint8_t sim_sta)
{
    // Keep track if we're registered on the network.
    radio_ready = (BP_API_RADIO_START_NOTIFY__REASON_PWRON_READY == notify_reason ||
                       BP_API_RADIO_START_NOTIFY__REASON_UNEXPECTED_REBOOT == notify_reason);

    sim_status = sim_sta;
    awaiting_reply = false;
}

void mobi_get_radio_status_reply(uint8_t seq_nr, uint8_t status, uint8_t radio_status, uint8_t sim_sta)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        radio_ready = (BP_API_GET_RADIO_STATUS__RADIO_PWRON_READY == radio_status);
        sim_status = sim_sta;
        state = radio_ready ? SET_SIM_ACCESS : WAIT_RADIO_READY;
        awaiting_reply = false;
    }
}

void mobi_set_sim_access_config_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = SET_APN_CONFIG;
        awaiting_reply = false;
    }
}

void mobi_set_apn_config_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = GET_MOBILE_CONN_STATUS;
        awaiting_reply = false;
    }
}

void mobi_get_mobile_conn_status_reply(uint8_t seq_nr, uint8_t status, uint8_t conn_ind, uint8_t link_tech,
                                              uint16_t net_lac, uint32_t net_cell_id, uint8_t signal_strength,
                                              uint8_t* net_mcc, uint8_t* net_mnc)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        network_registered = (BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_REGD_HOME == conn_ind ||
                              BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_REGD_FOREIGN == conn_ind);
        state = network_registered ? HTTP_CONN_INIT : WAIT_NETWORK_REGISTRATION;
        awaiting_reply = false;
    }
}

void mobi_mobile_conn_status_change_notify_post(uint8_t seq_nr, uint8_t conn_ind, uint8_t link_tech,
                                                       uint16_t net_lac, uint32_t net_cell_id)
{
    network_registered = (BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_REGD_HOME == conn_ind ||
                          BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_REGD_FOREIGN == conn_ind);
}

void ip_http_init_connection_reply(uint8_t seq_nr, uint8_t status, uint8_t conn_id)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        http_conn_id = conn_id;
        http_conn_status = HTTP_CONN_STATUS_OPEN;
        state = HTTP_WAIT_CONN;
        awaiting_reply = false;
    }
}

void ip_http_post_write_ready_notify_post(uint8_t seq_nr, uint8_t conn_id)
{
    if(http_conn_id == conn_id)
    {
        http_conn_status = HTTP_CONN_STATUS_ESTABLISHED;  // Can now send the POST payload.
    }
}

void ip_http_post_write_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        if (state == HTTP_POST_WRITE)
        {
            http_bytes_sent += http_bytes_written;
        }
        else if (state == HTTP_POST_SEND)
        {
            state = HTTP_WAIT_DATA;
        }
        awaiting_reply = false;
    }
}

void ip_http_rx_notify_request(uint8_t seq_nr, uint8_t conn_id, uint8_t data_len, uint8_t* data)
{
    // IMPORTANT: Reply to this request, or else the BitPipe will re-send this data chunk.
    bp_api_ip_http_rx_notify_ack(seq_nr);

    if (http_conn_id == conn_id)
    {
        // Do anything intelligent with the data
    }
}

void ip_connection_close_notify_post(uint8_t seq_nr, uint8_t conn_id, uint8_t reason)
{
    if(http_conn_id == conn_id)
    {
        http_conn_status = HTTP_CONN_STATUS_CLOSED;
        http_conn_id = BP_API_IP_INVALID_CONNECTION_ID;
    }
}

void uart_write(const uint8_t* data, uint16_t data_len)
{
    uint16_t written = 0;

    while(written < data_len)
    {
        if (SERIAL_PORT.availableForWrite() > 0)
        {
            SERIAL_PORT.write(data[written]);
            written++;
        }
    }
}

void setup()
{
    uint8_t status;                                 // Request execution status.

    digitalWrite(LED_PIN, LOW);
    pinMode(LED_PIN, OUTPUT);

    // Initialize Serial
    SERIAL_PORT.begin(SERIAL_SPEED, SERIAL_CONFIG);

    // Initialize SDK layer.
    init_params.app_callbacks       =   &callbacks;
    init_params.uart_write          =   uart_write;
    init_params.deserialize_buf     =   deser_buf;
    init_params.deserialize_buf_len =   sizeof(deser_buf);
    init_params.serialize_buf       =   ser_buf;
    init_params.serialize_buf_len   =   sizeof(ser_buf);

    status = bp_api_init(&init_params);
    if(status != 0)
    {
        digitalWrite(LED_PIN, HIGH);
        while(1);
    }

    http_conn_status = HTTP_CONN_STATUS_CLOSED;
    http_conn_id = BP_API_IP_INVALID_CONNECTION_ID;

    awaiting_reply = false;
    radio_ready = false;
    network_registered = false;
    state = POWER_UP;
}

void loop()
{
    uint8_t status;                                 // Request execution status.
    uint16_t rx_data_len;
    uint8_t uart_rx_data;
    static unsigned long power_up_ts = millis();

    digitalWrite(LED_PIN, millis() % 1000 < 50 ? HIGH : LOW);

    // Check if we've received data from the BitPipe, and process it.
    if (SERIAL_PORT.available() > 0)
    {
        uart_rx_data = (uint8_t)SERIAL_PORT.read();
        bp_api_process_incoming(&uart_rx_data, sizeof(uart_rx_data), &rx_data_len);
    }

    if(!awaiting_reply)
    {
        status = 0xFF;
        switch(state)
        {
            case POWER_UP:
                if ((millis() - power_up_ts) > 5000UL)
                {
                    state = SET_MAX_PAYLOAD_SIZE;
                }
                break;

            case SET_MAX_PAYLOAD_SIZE:
                // Configure Maximum message payload size.
                status = bp_api_sys_set_max_payload_sz(&seq_number, MAX_PAYLOAD_SIZE);
                // state is advance when receive reply from bitpipe
                break;

            case GET_HW_VERSION:
                // Get HW version.
                status = bp_api_devi_get_hw_version(&seq_number);
                break;

            case GET_FW_VERSION:
                // Get FW version.
                status = bp_api_devi_get_fw_version(&seq_number);
                break;

            case POWER_ON_RADIO:
                // Power-on Radio.
                status = bp_api_mobi_power_on_radio(&seq_number,
                                                    BP_API_PWRON_RADIO__PWRON_METHOD_DCDC,
                                                    BP_API_PWRON_RADIO__BACKOFF_DISABLED);
                break;

            case GET_RADIO_STATUS:
                // Get Radio Status.
                status = bp_api_mobi_get_radio_status(&seq_number);
                break;

            case WAIT_RADIO_READY:
                // Wait for radio to be ready.
                if(radio_ready)
                {
                    state = SET_SIM_ACCESS;
                }
                break;

            case SET_SIM_ACCESS:     // Check SIM card status.
                if(BP_API_RADIO_START_NOTIFY__SIM_NOT_INSERTED == sim_status)
                {
                    state = NO_SIM_DETECTED;
                }
                else if(BP_API_RADIO_START_NOTIFY__SIM_DENIED_NEED_PIN == sim_status)
                {
                    status = bp_api_mobi_set_sim_access_config(&seq_number, strlen(SIM_PIN), (uint8_t*)SIM_PIN);
                }
                else if(BP_API_RADIO_START_NOTIFY__SIM_DENIED_NEED_PUK == sim_status)
                {
                    status = bp_api_mobi_set_sim_access_config(&seq_number, strlen(SIM_PUK), (uint8_t*)SIM_PUK);
                }
                else    // SIM is OK.
                {
                    state = SET_APN_CONFIG;
                }
                break;

            case SET_APN_CONFIG:
                // Set APN Configuration (This doesn't need to be set every time).
                status = bp_api_mobi_set_apn_config(&seq_number,
                                                    strlen(APN_CONFIG_HOST), (uint8_t*)APN_CONFIG_HOST,
                                                    strlen(APN_CONFIG_USER), (uint8_t*)APN_CONFIG_USER,
                                                    sizeof(APN_CONFIG_PASS)-1, (uint8_t*)APN_CONFIG_PASS);
                break;

            case GET_MOBILE_CONN_STATUS:
                // Get Mobile connection status.
                status = bp_api_mobi_get_mobile_conn_status(&seq_number);
                break;

            case WAIT_NETWORK_REGISTRATION:
                // Wait for network registration.
                if(network_registered)
                {
                    state = HTTP_CONN_INIT;
                }
                break;

            case HTTP_CONN_INIT:
                // Initiate HTTP connection.
                status = bp_api_ip_http_init_connection(&seq_number,
                                                        BP_API_IP_HTTP_INIT_CONNECTION__SECURE_NO,
                                                        BP_API_IP_HTTP_INIT_CONNECTION__METHOD_POST,
                                                        strlen(HTTP_REMOTE_HOST), (uint8_t*)HTTP_REMOTE_HOST,
                                                        HTTP_REMOTE_PORT,
                                                        strlen(HTTP_PATH_QUERY), (uint8_t*)HTTP_PATH_QUERY,
                                                        strlen(HTTP_ADD_HEADERS), (uint8_t*)HTTP_ADD_HEADERS,
                                                        HTTP_POST_DATA_LEN);
                break;

            case HTTP_WAIT_CONN:
                if (http_conn_status == HTTP_CONN_STATUS_ESTABLISHED)
                {
                    state = HTTP_POST_WRITE;
                }
                else if (http_conn_status == HTTP_CONN_STATUS_CLOSED)
                {
                    state = DONE;
                }
                break;

            case HTTP_POST_WRITE:
                if (http_conn_status == HTTP_CONN_STATUS_CLOSED)
                {
                    state = HTTP_CONN_INIT;
                }
                else if (http_bytes_sent < HTTP_POST_DATA_LEN)
                {
                    http_bytes_written = (HTTP_POST_DATA_LEN - http_bytes_sent) > HTTP_TMP_BLOCK_LEN ?
                                            HTTP_TMP_BLOCK_LEN :
                                            (HTTP_POST_DATA_LEN - http_bytes_sent);
#ifdef ARDUINO_ARCH_AVR
                    for (int i = 0; i < http_bytes_written; i++)
                    {
                        http_tmp_buffer[i] = pgm_read_byte(&http_post_data[http_bytes_sent]);
                    }
#else
                    memcpy(http_tmp_buffer, &http_post_data[http_bytes_sent], http_bytes_written);
#endif
                    // Not needed to split POST in multiple send, can fit in one buffer of BitPipe
                    status = bp_api_ip_http_post_write(&seq_number, http_conn_id, BP_API_IP_HTTP_POST_WRITE__WRITEOP_WRITE_TO_BUFFER, http_bytes_written, (uint8_t*)http_tmp_buffer);
                }
                else
                {
                    state = HTTP_POST_SEND;
                }
                break;

            case HTTP_POST_SEND:
                if (http_conn_status == HTTP_CONN_STATUS_CLOSED)
                {
                    state = DONE;
                }
                else
                {
                    status = bp_api_ip_http_post_write(&seq_number, http_conn_id, BP_API_IP_HTTP_POST_WRITE__WRITEOP_WRITE_TO_BUFFER_AND_SEND_COMPLETE, 0, NULL);
                }
                break;

            case HTTP_WAIT_DATA:
                if (http_conn_status == HTTP_CONN_STATUS_CLOSED)
                {
                    state = DONE;
                }
                break;

            case DONE:
                while(1)
                {
                    digitalWrite(LED_PIN, HIGH);
                    delay(500);
                    digitalWrite(LED_PIN, LOW);
                    delay(500);
                }
                break;

            case NO_SIM_DETECTED:
                // No SIM is inserted, we can do nothing
                while (1)
                {
                    digitalWrite(LED_PIN, HIGH);
                    delay(200);
                    digitalWrite(LED_PIN, LOW);
                    delay(200);
                }
                break;
        }

        if (status == 0)
        {
            awaiting_reply = true;
            request_ts = millis();
        }
        else if (status != 0xFF)
        {
            while(1)
            {
                digitalWrite(LED_PIN, HIGH);
                delay(50);
                digitalWrite(LED_PIN, LOW);
                delay(50);
            }
        }
    }
    else
    {
        // Check for sent request timeout.
        if ((millis() - request_ts) > REQUEST_TIMEOUT)
        {
            // This will trigger a re-send of the request (with a different sequence number).
            awaiting_reply = false;
        }
    }
}

