/** ********************************************************************************************************************
 * @file        MQTT.ino
 * @author      Briowireless, Inc.
 * @date        2017-01-09
 * @brief       Main program file for MQTT Sample.
 * @version     API V1
 * @copyright   Briowireless, Inc., 2017, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/

// Include directives  -------------------------------------------------------------------------------------------------

#include <bp_api.h>

// Preprocessor definitions --------------------------------------------------------------------------------------------

#define LED_PIN                             13

#define MAX_PAYLOAD_SIZE                    260

#ifdef ARDUINO_SAM_ZERO
#define SERIAL_PORT                         Serial1             // Serial class instance on which the BitPipe is found.
#else
#define SERIAL_PORT                         Serial              // Serial class instance on which the BitPipe is found.
#endif
#if (F_CPU > 8000000L)
#define SERIAL_SPEED                        115200
#else
#define SERIAL_SPEED                        9600
#endif
#define SERIAL_CONFIG                       SERIAL_8N1

#define REQUEST_TIMEOUT                     (30*1000UL)         // 30 seconds
#define PUBLISH_TIMEOUT                     (10*1000UL)         // 10 seconds

#define SIM_PIN                             ""                  // When SIM card is locked using a PIN
#define SIM_PUK                             ""                  // When too many bad PINs were tried, the PIN Unlock Key is necessary.

#define APN_CONFIG_HOST                     ""                  // TODO: Enter you APN host address here
#define APN_CONFIG_USER                     ""
#define APN_CONFIG_PASS                     ""

// -------- MQTT: If you want a public MQTT server, check out cloudmqtt.com for trials.
#define MQTT_REMOTE_HOST                    ""                  // TODO: Enter the server host address here
#define MQTT_REMOTE_PORT                    1883               // TODO: Enter the server port here
#define MQTT_SECURE                         0
#define MQTT_CLEAN_SESSION                  1
#define MQTT_CLIENT_ID                      "MqttSample__Cli1"  // TODO: Choose a unique client ID
#define MQTT_USER                           ""                  // TODO: Enter your instance username here
#define MQTT_PASS                           ""                  // TODO: Enter your instance password here
#define MQTT_KEEPALIVE_IVAL                 300/5
#define MQTT_SUB_TOP1                       "outputs/#"
//#define MQTT_SUB_TOP2                       ""
//#define MQTT_SUB_TOP3                       ""
#define MQTT_PUBLISH_TOPIC_TEMPERATURE      "sensors/temperature"
#define MQTT_PUBLISH_TOPIC_HUMIDITY         "sensors/humidity"
#define MQTT_PUBLISH_QOS                    0
#define MQTT_PUBLISH_RETAIN_IND             0

#define MQTT_OUTPUT_RELAY_X_TOPIC           "outputs/relay"
#define MQTT_OUTPUT_PWM_X_TOPIC             "outputs/pwm"

#define MQTT_CONN_STATUS_CLOSED             0
#define MQTT_CONN_STATUS_OPEN               1
#define MQTT_CONN_STATUS_ESTABLISHED        2

#define OFF                                 0
#define ON                                  1

#define RELAY1_PIN                          4
#define RELAY2_PIN                          7
#define RELAY3_PIN                          8
#define RELAY4_PIN                          12
#define MAX_RELAY_OUTPUTS                   4

#define PWM1_PIN                            5
#define PWM2_PIN                            6
#define MAX_PWM_OUTPUTS                     2

typedef enum 
{
    POWER_UP,
    SET_MAX_PAYLOAD_SIZE,
    GET_HW_VERSION,
    GET_FW_VERSION,
    POWER_ON_RADIO,
    GET_RADIO_STATUS,
    WAIT_RADIO_READY,
    SET_SIM_ACCESS,
    SET_APN_CONFIG,
    GET_MOBILE_CONN_STATUS,
    WAIT_NETWORK_REGISTRATION,
    MQTT_CONN_INIT,
    MQTT_WAIT_CONNECT,
    MQTT_SUBSCRIBE_TOPICS,
    MQTT_CONNECTED,
    NO_SIM_DETECTED,
}State;

typedef enum 
{
    TEMPERATURE = 0,
    HUMIDITY,
    SENSORS_COUNT,
}Sensors;

// Local functions prototypes ------------------------------------------------------------------------------------------

// Callbacks
void sys_set_max_payload_sz_reply(uint8_t seq_nr, uint8_t status);
void devi_get_hw_version_reply(uint8_t seq_nr, uint8_t status, uint8_t hw_ver_len, uint8_t* hw_ver);
void devi_get_fw_version_reply(uint8_t seq_nr, uint8_t status, uint16_t boot_ver, uint8_t api_ver,
                               uint8_t bp_ver_len, uint8_t* bp_ver, uint8_t radio_ver_len, uint8_t* radio_ver);
void mobi_power_on_radio_reply(uint8_t seq_nr, uint8_t status);
void mobi_radio_start_notify_post(uint8_t seq_nr, uint8_t notify_reason, uint8_t sim_sta);
void mobi_get_radio_status_reply(uint8_t seq_nr, uint8_t status, uint8_t radio_status, uint8_t sim_sta);
void mobi_set_sim_access_config_reply(uint8_t seq_nr, uint8_t status);
void mobi_set_apn_config_reply(uint8_t seq_nr, uint8_t status);
void mobi_get_mobile_conn_status_reply(uint8_t seq_nr, uint8_t status, uint8_t conn_ind, uint8_t link_tech,
                                              uint16_t net_lac, uint32_t net_cell_id, uint8_t signal_strength,
                                              uint8_t* net_mcc, uint8_t* net_mnc);
void mobi_mobile_conn_status_change_notify_post(uint8_t seq_nr, uint8_t conn_ind, uint8_t link_tech,
                                                       uint16_t net_lac, uint32_t net_cell_id);
void mqtt_init_connection_reply(uint8_t seq_nr, uint8_t status, uint8_t conn_id);
void mqtt_connection_established_notify_post(uint8_t seq_nr, uint8_t conn_id);
void mqtt_subscribe_topics_reply(uint8_t seq_nr, uint8_t status, uint8_t sub_status);
void mqtt_publish_reply(uint8_t seq_nr, uint8_t status);
void mqtt_publication_rx_notify_request(uint8_t seq_nr, uint8_t conn_id, uint8_t qos, uint8_t retained,
                                               uint8_t complete, uint8_t topic_len, uint8_t* topic,
                                               uint8_t data_len, uint8_t* data);
void mqtt_disconnect_notify_post(uint8_t seq_nr, uint8_t conn_id, uint8_t reason);

void uart_write(const uint8_t* data, uint16_t data_len);

// Local variables  ----------------------------------------------------------------------------------------------------

// Buffers used by SDK.
uint8_t deser_buf[MAX_PAYLOAD_SIZE + MESSAGE_PROTOCOL_HEADER_SIZE + MESSAGE_PROTOCOL_FOOTER_SIZE];
uint8_t ser_buf[MAX_PAYLOAD_SIZE];

bp_api_init_params init_params;

State state;
mp_seq_nr seq_number;
bool awaiting_reply;
unsigned long request_ts;
bool radio_ready;
uint8_t sim_status;
bool network_registered;

unsigned long status_request_ts;

uint8_t mqtt_conn_id;
uint8_t mqtt_conn_status;

Sensors sensor_to_publish;

bool publish_in_progress;
unsigned long publish_ts;

int temperature;
bool temperature_changed;

int humidity;
bool humidity_changed;

bool relay_output_state[MAX_RELAY_OUTPUTS];
bool relay_output_changed[MAX_RELAY_OUTPUTS];

uint16_t pwm_output_value[MAX_PWM_OUTPUTS];
bool pwm_output_changed[MAX_PWM_OUTPUTS];

char tmp_data_buffer[5];

/* Unused callbacks can be set to NULL */
bp_api_callbacks callbacks  = {
    NULL,                                       // bp_api_sys_error_report_post
    sys_set_max_payload_sz_reply,               // bp_api_sys_set_max_payload_sz_reply
    NULL,                                       // bp_api_sys_set_uart_speed_reply

    devi_get_hw_version_reply,                  // bp_api_devi_get_hw_version_reply
    devi_get_fw_version_reply,                  // bp_api_devi_get_fw_version_reply
    NULL,                                       // bp_api_devi_get_product_name_reply
    NULL,                                       // bp_api_devi_get_vin_vio_voltage_reply
    NULL,                                       // bp_api_devi_get_junction_temp_reply
    NULL,                                       // bp_api_devi_set_low_power_reply
    NULL,                                       // bp_api_devi_get_latest_fw_version_reply
    NULL,                                       // bp_api_devi_update_fw_reply
    NULL,                                       // bp_api_devi_get_autonomous_mode_reply
    NULL,                                       // bp_api_devi_set_autonomous_mode_reply

    mobi_power_on_radio_reply,                  // bp_api_mobi_power_on_radio_reply
    NULL,                                       // bp_api_mobi_power_off_radio_reply
    mobi_radio_start_notify_post,               // bp_api_mobi_radio_start_notify_post
    mobi_get_radio_status_reply,                // bp_api_mobi_get_radio_status_reply
    NULL,                                       // bp_api_mobi_get_imei_reply
    NULL,                                       // bp_api_mobi_get_iccid_reply
    NULL,                                       // bp_api_mobi_get_imsi_reply
    NULL,                                       // bp_api_mobi_get_own_phone_number_reply
    NULL,                                       // bp_api_mobi_get_apn_config_reply
    mobi_set_apn_config_reply,                  // bp_api_mobi_set_apn_config_reply
    mobi_set_sim_access_config_reply,           // bp_api_mobi_set_sim_access_config_reply
    mobi_get_mobile_conn_status_reply,          // bp_api_mobi_get_mobile_conn_status_reply
    mobi_mobile_conn_status_change_notify_post, // bp_api_mobi_mobile_conn_status_change_notify_post
    NULL,                                       // bp_api_mobi_get_mobile_date_time_reply

    NULL,                                       // bp_api_ip_tcp_init_connection_reply
    NULL,                                       // bp_api_ip_tcp_connection_established_notify_post
    NULL,                                       // bp_api_ip_tcp_write_append_reply
    NULL,                                       // bp_api_ip_tcp_write_send_reply
    NULL,                                       // bp_api_ip_tcp_rx_notify_request
    NULL,                                       // bp_api_ip_http_init_connection_reply
    NULL,                                       // bp_api_ip_http_post_write_ready_notify_post
    NULL,                                       // bp_api_ip_http_post_write_reply
    NULL,                                       // bp_api_ip_http_rx_notify_request
    NULL,                                       // bp_api_ip_connection_close_reply
    NULL,                                       // bp_api_ip_connection_close_notify_post

    mqtt_init_connection_reply,                 // bp_api_mqtt_init_connection_reply
    mqtt_connection_established_notify_post,    // bp_api_mqtt_connection_established_notify_post
    mqtt_subscribe_topics_reply,                // bp_api_mqtt_subscribe_topics_reply
    NULL,                                       // bp_api_mqtt_unsubscribe_topics_reply
    mqtt_publish_reply,                         // bp_api_mqtt_publish_reply
    mqtt_publication_rx_notify_request,         // bp_api_mqtt_publication_rx_notify_request
    NULL,                                       // bp_api_mqtt_disconnect_reply
    mqtt_disconnect_notify_post,                // bp_api_mqtt_disconnect_notify_post

    NULL,                                       // bp_api_sms_setup_transmission_reply
    NULL,                                       // bp_api_sms_message_append_send_reply
    NULL,                                       // bp_api_sms_message_rx_notify_request
    NULL,                                       // bp_api_sms_message_part_notify_request
};

// Local functions  ----------------------------------------------------------------------------------------------------

unsigned int str2int(const char* str, size_t len)
{
    int i;
    unsigned int ret = 0;
    for(i = 0; i < len; ++i)
    {
        ret = (ret * 10) + (str[i] - '0');
    }
    return ret;
}

void sys_set_max_payload_sz_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = GET_HW_VERSION;
        awaiting_reply = false;
    }
}

void devi_get_hw_version_reply(uint8_t seq_nr, uint8_t status, uint8_t hw_ver_len, uint8_t* hw_ver)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = GET_FW_VERSION;
        awaiting_reply = false;
    }
}

void devi_get_fw_version_reply(uint8_t seq_nr, uint8_t status, uint16_t boot_ver, uint8_t api_ver,
                               uint8_t bp_ver_len, uint8_t* bp_ver, uint8_t radio_ver_len, uint8_t* radio_ver)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = POWER_ON_RADIO;
        awaiting_reply = false;
    }
}

void mobi_power_on_radio_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        (status == MSG_STATUS_CODE_SUCCESS || status == BP_API_PWRON_RADIO__STATUS_ALREADY_STARTED))
    {
        if (status == MSG_STATUS_CODE_SUCCESS)
        {
            radio_ready = false;
            state = GET_RADIO_STATUS;
        }
        else if (status == BP_API_PWRON_RADIO__STATUS_ALREADY_STARTED)
        {
            radio_ready = true;
            state = SET_SIM_ACCESS;
        }
        awaiting_reply = false;
    }
}

void mobi_radio_start_notify_post(uint8_t seq_nr, uint8_t notify_reason, uint8_t sim_sta)
{
    // Keep track if we're registered on the network.
    radio_ready = (BP_API_RADIO_START_NOTIFY__REASON_PWRON_READY == notify_reason ||
                       BP_API_RADIO_START_NOTIFY__REASON_UNEXPECTED_REBOOT == notify_reason);

    sim_status = sim_sta;
    awaiting_reply = false;
}

void mobi_get_radio_status_reply(uint8_t seq_nr, uint8_t status, uint8_t radio_status, uint8_t sim_sta)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        radio_ready = (BP_API_GET_RADIO_STATUS__RADIO_PWRON_READY == radio_status);

        sim_status = sim_sta;

        state = WAIT_RADIO_READY;
        status_request_ts = millis();
        awaiting_reply = false;
    }
}

void mobi_set_sim_access_config_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = SET_APN_CONFIG;
        awaiting_reply = false;
    }
}

void mobi_set_apn_config_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = GET_MOBILE_CONN_STATUS;
        awaiting_reply = false;
    }
}

void mobi_get_mobile_conn_status_reply(uint8_t seq_nr, uint8_t status, uint8_t conn_ind, uint8_t link_tech,
                                              uint16_t net_lac, uint32_t net_cell_id, uint8_t signal_strength,
                                              uint8_t* net_mcc, uint8_t* net_mnc)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        network_registered = (BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_REGD_HOME == conn_ind ||
                              BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_REGD_FOREIGN == conn_ind);
        state = WAIT_NETWORK_REGISTRATION;
        status_request_ts = millis();
        awaiting_reply = false;
    }
}

void mobi_mobile_conn_status_change_notify_post(uint8_t seq_nr, uint8_t conn_ind, uint8_t link_tech,
                                                       uint16_t net_lac, uint32_t net_cell_id)
{
    if (!network_registered)
    {
        // If we are not connected and just received a status change for mobile connection, re-ask mobile connection status to BitPipe.
        state = GET_MOBILE_CONN_STATUS;
    }
    else if (BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_NOT_CONNECTED == conn_ind ||
             BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_CAMPING == conn_ind)
    {
        network_registered = false;
    }
}

void mqtt_init_connection_reply(uint8_t seq_nr, uint8_t status, uint8_t conn_id)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        mqtt_conn_id = conn_id;
        mqtt_conn_status = MQTT_CONN_STATUS_OPEN;
        state = MQTT_WAIT_CONNECT;
        awaiting_reply = false;
    }
}

void mqtt_connection_established_notify_post(uint8_t seq_nr, uint8_t conn_id)
{
    if (mqtt_conn_id == conn_id)
    {
        mqtt_conn_status = MQTT_CONN_STATUS_ESTABLISHED;
    }
}

void mqtt_subscribe_topics_reply(uint8_t seq_nr, uint8_t status, uint8_t sub_status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = MQTT_CONNECTED;
        awaiting_reply = false;
    }
}

void mqtt_publish_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        (status == MSG_STATUS_CODE_SUCCESS || status == BP_API_MQTT_PUBLISH__CANNOT_PUBLISH_NOT_READY))
    {
        publish_in_progress = false;
        if (status == MSG_STATUS_CODE_SUCCESS)
        {
            switch(sensor_to_publish)
            {
                case TEMPERATURE:   temperature_changed = false;    break;
                case HUMIDITY:      humidity_changed = false;       break;
            }
            sensor_to_publish = (Sensors)((int)(sensor_to_publish + 1) % SENSORS_COUNT);
        }
    }
}

void mqtt_publication_rx_notify_request(uint8_t seq_nr, uint8_t conn_id, uint8_t qos, uint8_t retained,
                                               uint8_t complete, uint8_t topic_len, uint8_t* topic,
                                               uint8_t data_len, uint8_t* data)
{
    // IMPORTANT: Reply to this request, or else the BitPipe will re-send this data chunk.
    bp_api_mqtt_publication_rx_notify_ack(seq_nr);

    if (mqtt_conn_id == conn_id)
    {
        int data_value = str2int((const char*)data, data_len);

        if (strncmp((const char*)topic, MQTT_OUTPUT_RELAY_X_TOPIC, strlen(MQTT_OUTPUT_RELAY_X_TOPIC)) == 0)
        {
            uint8_t output = str2int((const char*)&topic[strlen(MQTT_OUTPUT_RELAY_X_TOPIC)], topic_len - strlen(MQTT_OUTPUT_RELAY_X_TOPIC));
            if (output < MAX_RELAY_OUTPUTS)
            {
                relay_output_state[output] = data_value > 0 ? true : false;
                relay_output_changed[output] = true;
            }
        }
        else if (strncmp((const char*)topic, MQTT_OUTPUT_PWM_X_TOPIC, strlen(MQTT_OUTPUT_PWM_X_TOPIC)) == 0)
        {
            uint8_t output = str2int((const char*)&topic[strlen(MQTT_OUTPUT_PWM_X_TOPIC)], topic_len - strlen(MQTT_OUTPUT_PWM_X_TOPIC));
            if (output < MAX_PWM_OUTPUTS)
            {
                pwm_output_value[output] = data_value;
                pwm_output_changed[output] = true;
            }
        }
    }
}

void mqtt_disconnect_notify_post(uint8_t seq_nr, uint8_t conn_id, uint8_t reason)
{
    if(mqtt_conn_id == conn_id)
    {
        mqtt_conn_status = MQTT_CONN_STATUS_CLOSED;
        mqtt_conn_id = BP_API_IP_INVALID_CONNECTION_ID;
    }
}

void uart_write(const uint8_t* data, uint16_t data_len)
{
    uint16_t written = 0;

    while(written < data_len)
    {
        if (SERIAL_PORT.availableForWrite() > 0)
        {
            SERIAL_PORT.write(data[written]);
            written++;
        }
    }
}

void output_relay(int output, bool state)
{
    int pin = -1;

    switch(output)
    {
        case 0: pin = RELAY1_PIN;   break;
        case 1: pin = RELAY2_PIN;   break;
        case 2: pin = RELAY3_PIN;   break;
        case 3: pin = RELAY4_PIN;   break;
    }

    if (pin != -1)
    {
        digitalWrite(pin, state != OFF ? HIGH : LOW);
        pinMode(pin, OUTPUT);
    }
}

void output_pwm(int output, uint8_t value)
{
    int pin = -1;

    switch(output)
    {
        case 0: pin = PWM1_PIN;   break;
        case 1: pin = PWM2_PIN;   break;
    }

    if (pin != -1)
    {
        analogWrite(pin, value);
        pinMode(pin, OUTPUT);
    }
}

void process_sensors()
{
    uint8_t status;

    if (!publish_in_progress)
    {
        switch(sensor_to_publish)
        {
            case TEMPERATURE:
                if (temperature_changed)
                {
                    sprintf(tmp_data_buffer, "%d", temperature);
                    status = bp_api_mqtt_publish(&seq_number, mqtt_conn_id, MQTT_PUBLISH_QOS, MQTT_PUBLISH_RETAIN_IND, 1,
                                                    strlen(MQTT_PUBLISH_TOPIC_TEMPERATURE), (uint8_t*)MQTT_PUBLISH_TOPIC_TEMPERATURE,
                                                    strlen(tmp_data_buffer), (uint8_t*)tmp_data_buffer);
                    if (status == 0)
                    {
                        publish_in_progress = true;
                        publish_ts = millis();
                    }
                }
                break;

            case HUMIDITY:
                if (humidity_changed)
                {
                    sprintf(tmp_data_buffer, "%d", humidity);
                    status = bp_api_mqtt_publish(&seq_number, mqtt_conn_id, MQTT_PUBLISH_QOS, MQTT_PUBLISH_RETAIN_IND, 1,
                                                    strlen(MQTT_PUBLISH_TOPIC_HUMIDITY), (uint8_t*)MQTT_PUBLISH_TOPIC_HUMIDITY,
                                                    strlen(tmp_data_buffer), (uint8_t*)tmp_data_buffer);
                    if (status == 0)
                    {
                        publish_in_progress = true;
                        publish_ts = millis();
                    }
                }
                break;
        }
    }
    else if ((millis() - publish_ts) > PUBLISH_TIMEOUT)
    {
        publish_in_progress = false;
        sensor_to_publish = (Sensors)((int)(sensor_to_publish + 1) % SENSORS_COUNT);
    }
}

void process_outputs()
{
    int i;

    for (i=0; i<MAX_RELAY_OUTPUTS; i++)
    {
        if (relay_output_changed[i])
        {
            // change relay output state
            output_relay(i, relay_output_state[i]);
            relay_output_changed[i] = false;
        }
    }

    for (i=0; i<MAX_PWM_OUTPUTS; i++)
    {
        if (pwm_output_changed[i])
        {
            // change pwm output value
            output_pwm(i, pwm_output_value[i]);
            pwm_output_changed[i] = false;
        }
    }
}

void setup()
{
    uint8_t status;                                     // Request execution status.
    int i;

    digitalWrite(LED_PIN, LOW);
    pinMode(LED_PIN, OUTPUT);

    // Initialize Serial
    SERIAL_PORT.begin(SERIAL_SPEED, SERIAL_CONFIG);

    // Initialize SDK layer.
    init_params.app_callbacks       =   &callbacks;
    init_params.uart_write =            &uart_write;
    init_params.deserialize_buf     =   deser_buf;
    init_params.deserialize_buf_len =   sizeof(deser_buf);
    init_params.serialize_buf       =   ser_buf;
    init_params.serialize_buf_len   =   sizeof(ser_buf);

    status = bp_api_init(&init_params);
    if(status != 0)
    {
        digitalWrite(LED_PIN, HIGH);
        while(1);
    }

    mqtt_conn_status = MQTT_CONN_STATUS_CLOSED;
    mqtt_conn_id = BP_API_IP_INVALID_CONNECTION_ID;

    awaiting_reply = false;
    radio_ready = false;
    network_registered = false;
    state = POWER_UP;

    temperature_changed = false;
    humidity_changed = false;
    for (i=0; i<MAX_RELAY_OUTPUTS; i++)
    {
        output_relay(i,OFF);
        relay_output_changed[i] = false;
    }

    for (i=0; i<MAX_PWM_OUTPUTS; i++)
    {
        output_pwm(i,OFF);
        pwm_output_changed[i] = false;
    }

    sensor_to_publish = TEMPERATURE;
    publish_in_progress = false;
}

void loop()
{
    uint8_t status;                                 // Request execution status.
    uint16_t rx_data_len;
    uint8_t uart_rx_data;
    static unsigned long read_ts;
    static unsigned long power_up_ts = millis();

    digitalWrite(LED_PIN, millis() % 1000 < 50 ? HIGH : LOW);

    // Update sensors each 10 seconds
    if ((millis() - read_ts) > 10000UL)
    {
        temperature = (rand() % 166) - 40;  // -40C to 125C
        temperature_changed = true;

        humidity = rand() % 101;            // 0% to 100%
        humidity_changed = true;

        read_ts = millis();
    }

    // Check if we've received data from the BitPipe, and process it.
    if (SERIAL_PORT.available() > 0)
    {
        uart_rx_data = (uint8_t)SERIAL_PORT.read();
        bp_api_process_incoming(&uart_rx_data, sizeof(uart_rx_data), &rx_data_len);
    }

    if(!awaiting_reply)
    {
        status = 0xFF;
        switch(state)
        {
            case POWER_UP:
                if ((millis() - power_up_ts) > 5000UL)
                {
                    state = SET_MAX_PAYLOAD_SIZE;
                }
                break;

            case SET_MAX_PAYLOAD_SIZE:
                // Configure Maximum message payload size.
                status = bp_api_sys_set_max_payload_sz(&seq_number, MAX_PAYLOAD_SIZE);
                // state is advance when receive reply from bitpipe
                break;

            case GET_HW_VERSION:
                // Get HW version.
                status = bp_api_devi_get_hw_version(&seq_number);
                break;

            case GET_FW_VERSION:
                // Get FW version.
                status = bp_api_devi_get_fw_version(&seq_number);
                break;

            case POWER_ON_RADIO:
                // Power-on Radio.
                status = bp_api_mobi_power_on_radio(&seq_number,
                                                    BP_API_PWRON_RADIO__PWRON_METHOD_DCDC,
                                                    BP_API_PWRON_RADIO__BACKOFF_DISABLED);
                break;

            case GET_RADIO_STATUS:
                // Get Radio Status.
                status = bp_api_mobi_get_radio_status(&seq_number);
                break;

            case WAIT_RADIO_READY:
                // Wait for radio to be ready.
                if(radio_ready)
                {
                    state = SET_SIM_ACCESS;
                }
                else if ((millis() - status_request_ts) > REQUEST_TIMEOUT)
                {
                    state = GET_RADIO_STATUS;
                }
                break;

            case SET_SIM_ACCESS:     // Check SIM card status.
                if(BP_API_RADIO_START_NOTIFY__SIM_NOT_INSERTED == sim_status)
                {
                    state = NO_SIM_DETECTED;
                }
                else if(BP_API_RADIO_START_NOTIFY__SIM_DENIED_NEED_PIN == sim_status)
                {
                    status = bp_api_mobi_set_sim_access_config(&seq_number, strlen(SIM_PIN), (uint8_t*)SIM_PIN);
                }
                else if(BP_API_RADIO_START_NOTIFY__SIM_DENIED_NEED_PUK == sim_status)
                {
                    status = bp_api_mobi_set_sim_access_config(&seq_number, strlen(SIM_PUK), (uint8_t*)SIM_PUK);
                }
                else    // SIM is OK.
                {
                    state = SET_APN_CONFIG;
                }
                break;

            case SET_APN_CONFIG:
                // Set APN Configuration (This doesn't need to be set every time).
                status = bp_api_mobi_set_apn_config(&seq_number,
                                                    strlen(APN_CONFIG_HOST), (uint8_t*)APN_CONFIG_HOST,
                                                    strlen(APN_CONFIG_USER), (uint8_t*)APN_CONFIG_USER,
                                                    strlen(APN_CONFIG_PASS), (uint8_t*)APN_CONFIG_PASS);
                break;

            case GET_MOBILE_CONN_STATUS:
                // Get Mobile connection status.
                status = bp_api_mobi_get_mobile_conn_status(&seq_number);
                break;

            case WAIT_NETWORK_REGISTRATION:
                // Wait for network registration.
                if(network_registered)
                {
                    state = MQTT_CONN_INIT;
                }
                // If no message received in 30 seconds, re-ask for status
                else if ((millis() - status_request_ts) > REQUEST_TIMEOUT)
                {
                    // This will trigger a re-send of the get mobile conn status (with a different sequence number).
                    state = GET_MOBILE_CONN_STATUS;
                }
                break;

            case MQTT_CONN_INIT:
                if(network_registered)
                {
                    // MQTT: Initiate connection.
                    status = bp_api_mqtt_init_connection(&seq_number, MQTT_SECURE, MQTT_CLEAN_SESSION,
                                                        strlen(MQTT_REMOTE_HOST), (uint8_t*)MQTT_REMOTE_HOST,
                                                        MQTT_REMOTE_PORT,
                                                        strlen(MQTT_CLIENT_ID), (uint8_t*)MQTT_CLIENT_ID,
                                                        strlen(MQTT_USER), (uint8_t*)MQTT_USER,
                                                        strlen(MQTT_PASS), (uint8_t*)MQTT_PASS,
                                                        MQTT_KEEPALIVE_IVAL);
                }
                else
                {
                    state = GET_MOBILE_CONN_STATUS;
                }
                break;

            case MQTT_WAIT_CONNECT:
                if (mqtt_conn_status == MQTT_CONN_STATUS_ESTABLISHED)
                {
                    state = MQTT_SUBSCRIBE_TOPICS;
                }
                else if (!network_registered)
                {
                    mqtt_conn_status = MQTT_CONN_STATUS_CLOSED;
                    mqtt_conn_id = BP_API_IP_INVALID_CONNECTION_ID;
                    state = GET_MOBILE_CONN_STATUS;
                }
                break;

            case MQTT_SUBSCRIBE_TOPICS:
                if(network_registered)
                {
                    // MQTT: Subscribe to topics.
                    status = bp_api_mqtt_subscribe_topics(&seq_number, mqtt_conn_id, 0, 0, 0,
#ifdef MQTT_SUB_TOP1
                                                          strlen(MQTT_SUB_TOP1), (uint8_t*)MQTT_SUB_TOP1,
#else
                                                          0, (uint8_t*)NULL,
#endif
#ifdef MQTT_SUB_TOP2
                                                          strlen(MQTT_SUB_TOP2), (uint8_t*)MQTT_SUB_TOP2,
#else
                                                          0, (uint8_t*)NULL,
#endif
#ifdef MQTT_SUB_TOP3
                                                          strlen(MQTT_SUB_TOP3), (uint8_t*)MQTT_SUB_TOP3
#else
                                                          0, (uint8_t*)NULL
#endif
                                                          );
                }
                else
                {
                    mqtt_conn_status = MQTT_CONN_STATUS_CLOSED;
                    mqtt_conn_id = BP_API_IP_INVALID_CONNECTION_ID;
                    state = GET_MOBILE_CONN_STATUS;
                }
                break;

            case MQTT_CONNECTED:
                if (!network_registered)
                {
                    mqtt_conn_status = MQTT_CONN_STATUS_CLOSED;
                    mqtt_conn_id = BP_API_IP_INVALID_CONNECTION_ID;
                    state = GET_MOBILE_CONN_STATUS;
                }
                else if (mqtt_conn_status == MQTT_CONN_STATUS_CLOSED)
                {
                    state = MQTT_CONN_INIT;
                }
                else
                {
                    process_sensors();
                    process_outputs();
                }
                break;

            case NO_SIM_DETECTED:
                // No SIM is inserted, we can do nothing
                while (1)
                {
                    digitalWrite(LED_PIN, HIGH);
                    delay(100);
                    digitalWrite(LED_PIN, LOW);
                    delay(100);
                }
                break;
        }

        if (status == 0)
        {
            awaiting_reply = true;
            request_ts = millis();
        }
    }
    else
    {
        // Check for sent request timeout.
        if ((millis() - request_ts) > REQUEST_TIMEOUT)
        {
            // This will trigger a re-send of the request (with a different sequence number).
            awaiting_reply = false;
        }
    }
}

