/** ********************************************************************************************************************
 * @file        MQTTClien.cpp
 * @author      Briowireless, Inc.
 * @date        2017-06-07
 * @brief       API layer for MQTT which simplifies the use of the API.
 * @version     API V1
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/

// Include directives  -------------------------------------------------------------------------------------------------

#include "MQTTClient.h"

// Global Variables ----------------------------------------------------------------------------------------------------
HardwareSerial *serial_port;

State state;
uint8_t status;
uint8_t mqtt_conn_id;
uint8_t awaiting_reply;
uint8_t radio_ready;
uint8_t mqtt_conn_status;
bool publish_in_progress;
bool network_registered;

unsigned long request_ts;
unsigned long status_request_ts;
unsigned long publish_ts;

uint16_t rx_data_len;
uint8_t uart_rx_data;
char tmp_data_buffer[5];

uint8_t sim_status;

sub_recall_fct subscribe_recall;

mp_seq_nr seq_number;

uint8_t deser_buf[MAX_PAYLOAD_SIZE + MESSAGE_PROTOCOL_HEADER_SIZE + MESSAGE_PROTOCOL_FOOTER_SIZE];
uint8_t ser_buf[MAX_PAYLOAD_SIZE];

bp_api_callbacks callbacks = {
    NULL,                                       // bp_api_sys_error_report_post
    sys_set_max_payload_sz_reply,               // bp_api_sys_set_max_payload_sz_reply
    NULL,                                       // bp_api_sys_set_uart_speed_reply

    devi_get_hw_version_reply,                  // bp_api_devi_get_hw_version_reply
    devi_get_fw_version_reply,                  // bp_api_devi_get_fw_version_reply
    NULL,                                       // bp_api_devi_get_product_name_reply
    NULL,                                       // bp_api_devi_get_vin_vio_voltage_reply
    NULL,                                       // bp_api_devi_get_junction_temp_reply
    NULL,                                       // bp_api_devi_set_low_power_reply
    NULL,                                       // bp_api_devi_get_latest_fw_version_reply
    NULL,                                       // bp_api_devi_update_fw_reply
    NULL,                                       // bp_api_devi_get_autonomous_mode_reply
    NULL,                                       // bp_api_devi_set_autonomous_mode_reply

    mobi_power_on_radio_reply,                  // bp_api_mobi_power_on_radio_reply
    NULL,                                       // bp_api_mobi_power_off_radio_reply
    mobi_radio_start_notify_post,               // bp_api_mobi_radio_start_notify_post
    mobi_get_radio_status_reply,                // bp_api_mobi_get_radio_status_reply
    NULL,                                       // bp_api_mobi_get_imei_reply
    NULL,                                       // bp_api_mobi_get_iccid_reply
    NULL,                                       // bp_api_mobi_get_imsi_reply
    NULL,                                       // bp_api_mobi_get_own_phone_number_reply
    NULL,                                       // bp_api_mobi_get_apn_config_reply
    mobi_set_apn_config_reply,                  // bp_api_mobi_set_apn_config_reply
    mobi_set_sim_access_config_reply,           // bp_api_mobi_set_sim_access_config_reply
    mobi_get_mobile_conn_status_reply,          // bp_api_mobi_get_mobile_conn_status_reply
    mobi_mobile_conn_status_change_notify_post, // bp_api_mobi_mobile_conn_status_change_notify_post
    NULL,                                       // bp_api_mobi_get_mobile_date_time_reply

    NULL,                                       // bp_api_ip_tcp_init_connection_reply
    NULL,                                       // bp_api_ip_tcp_connection_established_notify_post
    NULL,                                       // bp_api_ip_tcp_write_append_reply
    NULL,                                       // bp_api_ip_tcp_write_send_reply
    NULL,                                       // bp_api_ip_tcp_rx_notify_request
    NULL,                                       // bp_api_ip_http_init_connection_reply
    NULL,                                       // bp_api_ip_http_post_write_ready_notify_post
    NULL,                                       // bp_api_ip_http_post_write_reply
    NULL,                                       // bp_api_ip_http_rx_notify_request
    NULL,                                       // bp_api_ip_connection_close_reply
    NULL,                                       // bp_api_ip_connection_close_notify_post

    mqtt_init_connection_reply,                 // bp_api_mqtt_init_connection_reply
    mqtt_connection_established_notify_post,    // bp_api_mqtt_connection_established_notify_post
    mqtt_subscribe_topics_reply,                // bp_api_mqtt_subscribe_topics_reply
    mqtt_unsubscribe_topics_reply,                                       // bp_api_mqtt_unsubscribe_topics_reply
    mqtt_publish_reply,                         // bp_api_mqtt_publish_reply
    mqtt_publication_rx_notify_request,         // bp_api_mqtt_publication_rx_notify_request
    mqtt_disconnect_reply,                      // bp_api_mqtt_disconnect_reply
    mqtt_disconnect_notify_post,                // bp_api_mqtt_disconnect_notify_post

    NULL,                                       // bp_api_sms_setup_transmission_reply
    NULL,                                       // bp_api_sms_message_append_send_reply
    NULL,                                       // bp_api_sms_message_rx_notify_request
    NULL,                                       // bp_api_sms_message_part_notify_request
};

// Global function  -------------------------------------------------------------------------------------------

/** ********************************************************************************************************************
 * @brief  MQTTClient Builder
 *
 * @param  client_id       Client identifier of the MQTT CONNECT operation.
 * @param  serial_port_lc  Arduino's Serial port used to communicat with the BitPipe
 * @param  apn_config_host APN's name used to establish data links with the mobile operator
 */
MQTTClient::MQTTClient(uint8_t *client_id, HardwareSerial *serial_port_lc, uint8_t *apn_config_host)
{

    this->client_id = client_id;
    serial_port = serial_port_lc;
    this->apn_config_host = apn_config_host;

    this->apn_config_user = NULL;
    this->apn_config_pass = NULL;

    this->sim_pin = NULL;
    this->sim_puk = NULL;
}

/** ********************************************************************************************************************
 * @brief  MQTTClient Builder
 *
 * @param  client_id        Client identifier of the MQTT CONNECT operation.
 * @param  serial_port_lc   Arduino's Serial port used to communicat with the BitPipe
 * @param  apn_config_host  APN's name used to establish data links with the mobile operator
 * @param  apn_config_user  APN's username
 * @param  apn_config_pass  APN's password
 * @param  sim_pin          SIM card's PIN
 * @param  sim_puk          SIM card's PUK
 */
MQTTClient::MQTTClient(uint8_t *client_id, HardwareSerial *serial_port_lc, uint8_t *apn_config_host,
                       uint8_t *apn_config_user, uint8_t *,
                       uint8_t *sim_pin, uint8_t *sim_puk)
{
    this->client_id = client_id;
    serial_port = serial_port_lc;
    this->apn_config_host = apn_config_host;

    this->apn_config_user = apn_config_user;
    this->apn_config_pass = apn_config_pass;

    this->sim_pin = sim_pin;
    this->sim_puk = sim_puk;
}

/** ********************************************************************************************************************
 * @brief  Initiates the MQTTClient's parameters
 *         Initiates the Arduino's serial port
 *
 * @param  serial_speed     Arduino's serial BAUD rate, must match the BitPipe's speed
 * @param  serial_config    Arduino's serial config
 * @param  f                User defined callback function that will be called when receiving data from a subscribed topic
 */
void MQTTClient::setup(long serial_speed, uint8_t serial_config, void (*f)(uint8_t* topic, uint8_t topic_len,
                                                                             uint8_t* data, uint8_t data_len,
                                                                             uint8_t qos, uint8_t retained,
                                                                             uint8_t complete))
{
    // Initialize Serial

    serial_port->begin(serial_speed, serial_config);

    // Initialize SDK layer.
    init_params.app_callbacks = &callbacks;
    init_params.uart_write = &uart_write;
    init_params.deserialize_buf = deser_buf;
    init_params.deserialize_buf_len = sizeof(deser_buf);
    init_params.serialize_buf = ser_buf;
    init_params.serialize_buf_len = sizeof(ser_buf);

    status = bp_api_init(&init_params);

    subscribe_recall = f;

    mqtt_conn_status = MQTT_CONN_STATUS_CLOSED;
    mqtt_conn_id = BP_API_IP_INVALID_CONNECTION_ID;

    awaiting_reply = false;
    radio_ready = false;
    network_registered = false;
    state = POWER_UP;

    publish_in_progress = false;

    this->user_connect = false;
}

/** ********************************************************************************************************************
 * @brief  Sets the parameters required to connect
 *         The connect action is done in the main loop
 *
 * @param  host_lc    Remote MQTT server hostname, or IP address.
 * @param  port_lc    MQTT remote port to connect to. If 0, 1883 or 8883 will be automatically chosen depending on the secure parameter.
 */
void MQTTClient::connect(uint8_t *host_lc, uint16_t port_lc)
{
    this->user_connect = true;

    this->host = host_lc;
    this->port = port_lc;

    this->user = NULL;
    this->pass = NULL;

    this->clean_session = BP_API_MQTT_CLEAN_SESSION_DEFAULT;
    this->secure = BP_API_MQTT_SECURE_DEFAULT;
    this->ka_ival_5_secs = BP_API_MQTT_KEEP_ALIVE_DEFAULT;
}

/** ********************************************************************************************************************
 * @brief  Sets the parameters required to connect
 *         The connect action is done in the main loop
 *
 * @param  host_lc    Remote MQTT server hostname, or IP address.
 * @param  port_lc    MQTT remote port to connect to.
 * @param  user       Username for authentication.
 * @param  pass       Password for authentication.
 */
void MQTTClient::connect(uint8_t *host_lc, uint16_t port_lc,
                            uint8_t clean_session_lc, uint8_t secure_lc, uint8_t ka_ival_5_secs_lc)
{
    this->user_connect = true;

    this->host = host_lc;
    this->port = port_lc;

    this->user = NULL;
    this->pass = NULL;

    this->clean_session = clean_session_lc;
    this->secure = secure_lc;
    this->ka_ival_5_secs = ka_ival_5_secs_lc;
}

/** ********************************************************************************************************************
 * @brief  Sets the parameters required to connect
 *         The connect action is done in the main loop
 *
 * @param  host_lc          Remote MQTT server hostname, or IP address.
 * @param  port_lc          MQTT remote port to connect to.
 * @param  user             Username for authentication.
 * @param  pass             Password for authentication.
 * @param  secure           Specifies if the MQTT connection should be encrypted using SSL/TLS. Values: [0,1]
 * @param  clean_session    Specifies whether we're starting a clean session, or resuming an exiting one. Values: [0,1]
 * @param  ka_ival_5_secs   Keep-Alive interval (in units of 5 seconds). Values: [0 (disabled), 60-240]
 */
void MQTTClient::connect(uint8_t *host_lc, uint16_t port_lc,
                            uint8_t *user_lc, uint8_t *pass_lc,
                            uint8_t clean_session_lc, uint8_t secure_lc, uint8_t ka_ival_5_secs_lc)
{
    this->user_connect = true;

    this->host = host_lc;
    this->port = port_lc;

    this->user = user_lc;
    this->pass = pass_lc;

    this->clean_session = clean_session_lc;
    this->secure = secure_lc;
    this->ka_ival_5_secs = ka_ival_5_secs_lc;
}

/** ********************************************************************************************************************
 * @brief  Sets the user_connect parameter to false.
 *         The disconnect action is done in the main loop
 */
void MQTTClient::disconnect()
{
    this->user_connect = false;
}

/** ********************************************************************************************************************
 * @brief  Sets the parameters required to subscribe
 *         The subscribe action is done in the main loop
 *
 * @param  topic    Topic to subscribe to
 */
void MQTTClient::subscribe(uint8_t *topic)
{
    if (state == MQTT_CONNECTED)
    {
		if (!publish_in_progress)
		{
			state = MQTT_SUBSCRIBE_TOPICS;

			this->subscribe_topic_1 = topic;
			strcpy ((char*) this->subscribe_topic_2, "");
			strcpy ((char*)this->subscribe_topic_3, "");

			this->subscribe_qos_1 = 0;
			this->subscribe_qos_2 = 0;
			this->subscribe_qos_3 = 0;
		}
    }
}

/** ********************************************************************************************************************
 * @brief  Sets the parameters required to subscribe
 *         The subscribe action is done in the main loop
 *
 * @param  topic    Topic to subscribe to
 * @param  qos      The QOS under which this topic is subscribed. Values: [0,1]
 */
void MQTTClient::subscribe(uint8_t *topic, uint8_t qos)
{
    if (state == MQTT_CONNECTED)
    {
		if (!publish_in_progress)
		{
			state = MQTT_SUBSCRIBE_TOPICS;

			this->subscribe_topic_1 = topic;
			strcpy ((char*) this->subscribe_topic_2, "");
			strcpy ((char*)this->subscribe_topic_3, "");

			this->subscribe_qos_1 = qos;
			this->subscribe_qos_2 = 0;
			this->subscribe_qos_3 = 0;
		}
    }
}

/** ********************************************************************************************************************
 * @brief  Sets the parameters required to subscribe to 1, 2 or 3 topics
 *         The subscribe action is done in the main loop
 *
 * @param  topic1   First topic to subscribe to
 * @param  qos1     The QOS under which the first topic is subscribed. Values: [0,1]
 * @param  topic2   Second topic to subscribe to
 * @param  qos2     The QOS under which the second topic is subscribed. Values: [0,1]
 * @param  topic3   Third topic to subscribe to
 * @param  qos 3    The QOS under which the third topic is subscribed. Values: [0,1]
 */
void MQTTClient::subscribe(uint8_t *topic1, uint8_t qos1, uint8_t *topic2, uint8_t qos2, uint8_t *topic3, uint8_t qos3)
{
    if (state == MQTT_CONNECTED)
    {
		if (!publish_in_progress)
		{
			state = MQTT_SUBSCRIBE_TOPICS;

			this->subscribe_topic_1 = topic1;
			this->subscribe_topic_2 = topic2;
			this->subscribe_topic_3 = topic3;

			this->subscribe_qos_1 = qos1;
			this->subscribe_qos_2 = qos2;
			this->subscribe_qos_3 = qos3;
		}
    }
}

/** ********************************************************************************************************************
 * @brief  Sets the parameters required to unsubscribe
 *         The unsubscribe action is done in the main loop
 *
 * @param  topic    Topic to unsubscribe from
 */
void MQTTClient::unsubscribe(uint8_t *topic)
{
    if (state == MQTT_CONNECTED)
    {
        state = MQTT_UNSUBSCRIBE_TOPICS;

        this->subscribe_topic_1 = topic;
		strcpy ((char*) this->subscribe_topic_2, "");
        strcpy ((char*)this->subscribe_topic_3, "");

        this->subscribe_qos_1 = 0;
        this->subscribe_qos_2 = 0;
        this->subscribe_qos_3 = 0;
    }


}

/** ********************************************************************************************************************
 * @brief  Sets the parameters required to unsubscribe from 1, 2 or 3 topics
 *         The unsubscribe action is done in the main loop
 *
 * @param  topic1   First topic to unsubscribe from
 * @param  topic2   Second topic to unsubscribe from
 * @param  topic3   Third topic to unsubscribe from
 */
 void MQTTClient::unsubscribe(uint8_t *topic1, uint8_t *topic2, uint8_t *topic3)
{
    if (state == MQTT_CONNECTED)
    {
        state = MQTT_UNSUBSCRIBE_TOPICS;

        this->subscribe_topic_1 = topic1;
        this->subscribe_topic_2 = topic2;
        this->subscribe_topic_3 = topic3;

        this->subscribe_qos_1 = 0;
        this->subscribe_qos_2 = 0;
        this->subscribe_qos_3 = 0;
    }


}

/** ********************************************************************************************************************
 * @brief  Sets the parameters required to publish
 *         The publish action is done in the main loop
 *         Does nothing if the loop is not already idle or if it is already publishing.
 *
 * @param  topic_lc   Topic, under which this message will be published.
 * @param  data_lc    Message data.
 */
void MQTTClient::publish(uint8_t *topic_lc, uint8_t *data_lc)
{
    if (state == MQTT_CONNECTED)
    {
        if (!publish_in_progress)
        {
            state = MQTT_PUBLISH;

            this->publish_topic = topic_lc;
            this->publish_data = data_lc;
            this->publish_qos = BP_API_MQTT_QOS_DEFAULT;
            this->publish_retained = BP_API_MQTT_RETAINED_DEFAULT;
        }
    }
}

/** ********************************************************************************************************************
 * @brief  Sets the parameters required to publish
 *         The publish action is done in the main loop
 *         Does nothing if the loop is not already idle or if it is already publishing.
 *
 * @param  topic_lc   Topic, under which this message will be published.
 * @param  data_lc    Message data.
 * @param  qos        The QOS under which this message is published. Values: [0,1]
 * @param  retained   Specifies whether the message is marked as retained. Values: [0,1]
 */
void MQTTClient::publish(uint8_t *topic_lc, uint8_t *data_lc, uint8_t qos, uint8_t retained)
{
    if (state == MQTT_CONNECTED)
    {
        if (!publish_in_progress)
        {
            state = MQTT_PUBLISH;

            this->publish_topic = topic_lc;
            this->publish_data = data_lc;
            this->publish_qos = qos;
            this->publish_retained = retained;
        }
    }
}

/** ********************************************************************************************************************
 * @brief  Returns the user_connect state.
 *
 * @retval bool  returns the state of the user connect switch
 */
bool MQTTClient::isconnected()
{
    return user_connect;
}

/** ********************************************************************************************************************
 * @brief  Main MQTT loop
 *
 *         Every action is done here. The communication process with the BitPipe is done here.
 *         The connection/reconnection process is also done here.
 *
 * @note   Only one action may be done at a time. This means that you cannot ask for two publishes
 *         on the same Arduino loop.
 *
 */
void MQTTClient::loop()
{

    // Check if we've received data from the BitPipe, and process it.
    if (serial_port->available() > 0)
    {
        uart_rx_data = (uint8_t)serial_port->read();
        bp_api_process_incoming(&uart_rx_data, sizeof(uart_rx_data), &rx_data_len);
    }

    if (!awaiting_reply)
    {
        status = 0xFF;
        switch (state)
        {
            case POWER_UP:

                if (millis() > 5000UL)
                {
                    state = SET_MAX_PAYLOAD_SIZE;
                }
                break;

            case SET_MAX_PAYLOAD_SIZE:

                // Configure Maximum message payload size.
                status = bp_api_sys_set_max_payload_sz(&seq_number, MAX_PAYLOAD_SIZE);
                // state is advance when receive reply from bitpipe
                break;

            case GET_HW_VERSION:

                // Get HW version.
                status = bp_api_devi_get_hw_version(&seq_number);
                break;

            case GET_FW_VERSION:

                // Get FW version.
                status = bp_api_devi_get_fw_version(&seq_number);
                break;

            case POWER_ON_RADIO:

                // Power-on Radio.
                status = bp_api_mobi_power_on_radio(&seq_number,
                                                    BP_API_PWRON_RADIO__PWRON_METHOD_DCDC,
                                                    BP_API_PWRON_RADIO__BACKOFF_DISABLED);
                break;

            case GET_RADIO_STATUS:

                // Get Radio Status.
                status = bp_api_mobi_get_radio_status(&seq_number);
                break;

            case WAIT_RADIO_READY:

                // Wait for radio to be ready.
                if (radio_ready)
                {
                    state = SET_SIM_ACCESS;
                }
                else if ((millis() - status_request_ts) > REQUEST_TIMEOUT)
                {
                    state = GET_RADIO_STATUS;
                }
                break;

            case SET_SIM_ACCESS: // Check SIM card status.

                if (BP_API_RADIO_START_NOTIFY__SIM_NOT_INSERTED == sim_status)
                {
                    state = NO_SIM_DETECTED;
                }
                else if (BP_API_RADIO_START_NOTIFY__SIM_DENIED_NEED_PIN == sim_status)
                {
                    status = bp_api_mobi_set_sim_access_config(&seq_number, strlen((const char *)this->sim_pin), (uint8_t *)this->sim_pin);
                }
                else if (BP_API_RADIO_START_NOTIFY__SIM_DENIED_NEED_PUK == sim_status)
                {
                    status = bp_api_mobi_set_sim_access_config(&seq_number, strlen((const char *)this->sim_puk), (uint8_t *)this->sim_puk);
                }
                else // SIM is OK.
                {
                    state = SET_APN_CONFIG;
                }
                break;

            case SET_APN_CONFIG:

                // Set APN Configuration (This doesn't need to be set every time).
                status = bp_api_mobi_set_apn_config(&seq_number,
                                                    strlen((const char *)this->apn_config_host), (uint8_t *)this->apn_config_host,
                                                    strlen((const char *)this->apn_config_user), (uint8_t *)this->apn_config_user,
                                                    strlen((const char *)this->apn_config_pass), (uint8_t *)this->apn_config_pass);
                break;

            case GET_MOBILE_CONN_STATUS:

                // Get Mobile connection status.
                status = bp_api_mobi_get_mobile_conn_status(&seq_number);
                break;

            case WAIT_NETWORK_REGISTRATION:

                // Wait for network registration.
                if (network_registered)
                {
                    state = MQTT_WAIT_USER_CONNECT;
                }
                // If no message received in 30 seconds, re-ask for status
                else if ((millis() - status_request_ts) > REQUEST_TIMEOUT)
                {
                    // This will trigger a re-send of the get mobile conn status (with a different sequence number).
                    state = GET_MOBILE_CONN_STATUS;
                }
                break;

            case MQTT_WAIT_USER_CONNECT:

                if (network_registered)
                {
                    if (this->user_connect)
                    {
                        state = MQTT_CONN_INIT;
                    }
                    //Waiting for the user to connect
                }
                else
                {
                    state = GET_MOBILE_CONN_STATUS;
                }
            break;


            case MQTT_CONN_INIT:

                if (network_registered)
                {

                    // MQTT: Initiate connection.
                    status = bp_api_mqtt_init_connection(&seq_number, this->secure, this->clean_session,
                                                        strlen((const char *)this->host), (uint8_t *)this->host,
                                                        this->port,
                                                        strlen((const char *)this->client_id), (uint8_t *)this->client_id,
                                                        strlen((const char *)this->user), (uint8_t *)this->user,
                                                        strlen((const char *)this->pass), (uint8_t *)this->pass,
                                                        this->ka_ival_5_secs);
                }
                else
                {
                    state = GET_MOBILE_CONN_STATUS;
                }
                break;

            case MQTT_WAIT_CONNECT:

                if (mqtt_conn_status == MQTT_CONN_STATUS_ESTABLISHED)
                {
                    state = MQTT_CONNECTED;
                }
                else if (!network_registered)
                {
                    mqtt_conn_status = MQTT_CONN_STATUS_CLOSED;
                    mqtt_conn_id = BP_API_IP_INVALID_CONNECTION_ID;
                    state = GET_MOBILE_CONN_STATUS;
                }
                break;



            case MQTT_CONNECTED:

                if(!user_connect)
                {
                    state = MQTT_DISCONNECT;
                }
                else if (!network_registered)
                {
                    mqtt_conn_status = MQTT_CONN_STATUS_CLOSED;
                    mqtt_conn_id = BP_API_IP_INVALID_CONNECTION_ID;
                    state = GET_MOBILE_CONN_STATUS;
                }
                else if (mqtt_conn_status == MQTT_CONN_STATUS_CLOSED)
                {
                    state = MQTT_CONN_INIT;
                }
            break;

            case MQTT_SUBSCRIBE_TOPICS:

                if (network_registered)
                {
                    // MQTT: Subscribe to topics.
                    status = bp_api_mqtt_subscribe_topics(&seq_number, mqtt_conn_id,
                                                        this->subscribe_qos_1, this->subscribe_qos_2, this->subscribe_qos_3,
                                                        strlen((const char*)this->subscribe_topic_1), (uint8_t*)this->subscribe_topic_1,
                                                        strlen((const char*)this->subscribe_topic_2), (uint8_t*)this->subscribe_topic_2,
                                                        strlen((const char*)this->subscribe_topic_3), (uint8_t*)this->subscribe_topic_3
                                                        );
                }
                else
                {
                    mqtt_conn_status = MQTT_CONN_STATUS_CLOSED;
                    mqtt_conn_id = BP_API_IP_INVALID_CONNECTION_ID;
                    state = GET_MOBILE_CONN_STATUS;
                }
            break;

            case MQTT_UNSUBSCRIBE_TOPICS:

                if (network_registered)
                {
                    // MQTT: Unsubscribe to topics.
                    status = bp_api_mqtt_unsubscribe_topics(&seq_number, mqtt_conn_id,
                                                        strlen((const char *)this->subscribe_topic_1), (uint8_t*)this->subscribe_topic_1,
                                                        strlen((const char *)this->subscribe_topic_2), (uint8_t*)this->subscribe_topic_2,
                                                        strlen((const char *)this->subscribe_topic_3), (uint8_t*)this->subscribe_topic_3
                                                        );
                }
                else
                {
                    mqtt_conn_status = MQTT_CONN_STATUS_CLOSED;
                    mqtt_conn_id = BP_API_IP_INVALID_CONNECTION_ID;
                    state = GET_MOBILE_CONN_STATUS;
                }
            break;

            case MQTT_PUBLISH:

                uint8_t status_publish;

                if (!publish_in_progress)
                {
                    sprintf(tmp_data_buffer, "%d", this->publish_data);
                    status_publish = bp_api_mqtt_publish(&seq_number, mqtt_conn_id, this->publish_qos, this->publish_retained, 1,
                                                         strlen((const char *)this->publish_topic), (uint8_t *)this->publish_topic,
                                                         strlen((const char *)tmp_data_buffer), (uint8_t *)tmp_data_buffer);
                    if (status_publish == 0)
                    {
                        publish_in_progress = true;
                        publish_ts = millis();
                    }
                }
                else if ((millis() - publish_ts) > PUBLISH_TIMEOUT)
                {
                    publish_in_progress = false;
                    state = MQTT_CONNECTED;
                }
            break;

            case MQTT_DISCONNECT:

                if (network_registered)
                {
                    status = bp_api_mqtt_disconnect(&seq_number, mqtt_conn_id);
                }
                else
                {
                    state = GET_MOBILE_CONN_STATUS;
                }
            break;

            case MQTT_WAIT_DISCONNECT:

                if (mqtt_conn_status == MQTT_CONN_STATUS_CLOSED)
                {
                    state = GET_MOBILE_CONN_STATUS;
                }
                else if (!network_registered)
                {
                    mqtt_conn_status = MQTT_CONN_STATUS_CLOSED;
                    mqtt_conn_id = BP_API_IP_INVALID_CONNECTION_ID;
                    state = GET_MOBILE_CONN_STATUS;
                }
            break;

            case NO_SIM_DETECTED:

            // No SIM is inserted, we can do nothing
            while (1)
            {

            }
            break;
        }

        if (status == 0)
        {
            awaiting_reply = true;
            request_ts = millis();
        }
    }
    else
    {
        // Check for sent request timeout.
        if ((millis() - request_ts) > REQUEST_TIMEOUT)
        {
            // This will trigger a re-send of the request (with a different sequence number).
            awaiting_reply = false;
        }
    }
}

// Callbacks --------------------------------------------------------------------------------------------

void sys_set_max_payload_sz_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = GET_HW_VERSION;
        awaiting_reply = false;
    }
}

void devi_get_hw_version_reply(uint8_t seq_nr, uint8_t status, uint8_t hw_ver_len, uint8_t *hw_ver)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = GET_FW_VERSION;
        awaiting_reply = false;
    }
}

void devi_get_fw_version_reply(uint8_t seq_nr, uint8_t status, uint16_t boot_ver, uint8_t api_ver,
                               uint8_t bp_ver_len, uint8_t *bp_ver, uint8_t radio_ver_len, uint8_t *radio_ver)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = POWER_ON_RADIO;
        awaiting_reply = false;
    }
}

void mobi_power_on_radio_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        (status == MSG_STATUS_CODE_SUCCESS || status == BP_API_PWRON_RADIO__STATUS_ALREADY_STARTED))
    {
        if (status == MSG_STATUS_CODE_SUCCESS)
        {
            radio_ready = false;
            state = GET_RADIO_STATUS;
        }
        else if (status == BP_API_PWRON_RADIO__STATUS_ALREADY_STARTED)
        {
            radio_ready = true;
            state = SET_SIM_ACCESS;
        }
        awaiting_reply = false;
    }
}

void mobi_radio_start_notify_post(uint8_t seq_nr, uint8_t notify_reason, uint8_t sim_sta)
{
    // Keep track if we're registered on the network.
    radio_ready = (BP_API_RADIO_START_NOTIFY__REASON_PWRON_READY == notify_reason ||
                   BP_API_RADIO_START_NOTIFY__REASON_UNEXPECTED_REBOOT == notify_reason);

    sim_status = sim_sta;
    awaiting_reply = false;
}

void mobi_get_radio_status_reply(uint8_t seq_nr, uint8_t status, uint8_t radio_status, uint8_t sim_sta)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        radio_ready = (BP_API_GET_RADIO_STATUS__RADIO_PWRON_READY == radio_status);

        sim_status = sim_sta;

        state = WAIT_RADIO_READY;
        status_request_ts = millis();
        awaiting_reply = false;
    }
}

void mobi_set_sim_access_config_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = SET_APN_CONFIG;
        awaiting_reply = false;
    }
}

void mobi_set_apn_config_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = GET_MOBILE_CONN_STATUS;
        awaiting_reply = false;
    }
}

void mobi_get_mobile_conn_status_reply(uint8_t seq_nr, uint8_t status, uint8_t conn_ind, uint8_t link_tech,
                                       uint16_t net_lac, uint32_t net_cell_id, uint8_t signal_strength,
                                       uint8_t *net_mcc, uint8_t *net_mnc)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        network_registered = (BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_REGD_HOME == conn_ind ||
                              BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_REGD_FOREIGN == conn_ind);
        state = WAIT_NETWORK_REGISTRATION;
        status_request_ts = millis();
        awaiting_reply = false;
    }
}

void mobi_mobile_conn_status_change_notify_post(uint8_t seq_nr, uint8_t conn_ind, uint8_t link_tech,
                                                uint16_t net_lac, uint32_t net_cell_id)
{
    if (!network_registered)
    {
        // If we are not connected and just received a status change for mobile connection, re-ask mobile connection status to BitPipe.
        state = GET_MOBILE_CONN_STATUS;
    }
    else if (BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_NOT_CONNECTED == conn_ind ||
             BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_CAMPING == conn_ind)
    {
        network_registered = false;
    }
}

void mqtt_init_connection_reply(uint8_t seq_nr, uint8_t status, uint8_t conn_id)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        mqtt_conn_id = conn_id;
        mqtt_conn_status = MQTT_CONN_STATUS_OPEN;
        state = MQTT_WAIT_CONNECT;
        awaiting_reply = false;
    }
}

void mqtt_connection_established_notify_post(uint8_t seq_nr, uint8_t conn_id)
{
    if (mqtt_conn_id == conn_id)
    {
        mqtt_conn_status = MQTT_CONN_STATUS_ESTABLISHED;
    }
}

void mqtt_subscribe_topics_reply(uint8_t seq_nr, uint8_t status, uint8_t sub_status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = MQTT_CONNECTED;
        awaiting_reply = false;
    }
}

void mqtt_unsubscribe_topics_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = MQTT_CONNECTED;
        awaiting_reply = false;
    }
}

void mqtt_publish_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        (status == MSG_STATUS_CODE_SUCCESS || status == BP_API_MQTT_PUBLISH__CANNOT_PUBLISH_NOT_READY))
    {
        publish_in_progress = false;
        state = MQTT_CONNECTED;
    }
}

void mqtt_publication_rx_notify_request(uint8_t seq_nr, uint8_t conn_id, uint8_t qos, uint8_t retained,
                                        uint8_t complete, uint8_t topic_len, uint8_t *topic,
                                        uint8_t data_len, uint8_t *data)
{
    // IMPORTANT: Reply to this request, or else the BitPipe will re-send this data chunk.
    bp_api_mqtt_publication_rx_notify_ack(seq_nr);

    if (mqtt_conn_id == conn_id)
    {
        subscribe_recall(topic, topic_len, data, data_len, qos, retained, complete);
    }
}

void mqtt_disconnect_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        mqtt_conn_id = BP_API_IP_INVALID_CONNECTION_ID;
        mqtt_conn_status = MQTT_CONN_STATUS_CLOSED;
        state = MQTT_WAIT_DISCONNECT;
        awaiting_reply = false;
    }
}

void mqtt_disconnect_notify_post(uint8_t seq_nr, uint8_t conn_id, uint8_t reason)
{
    if (mqtt_conn_id == conn_id)
    {
        mqtt_conn_status = MQTT_CONN_STATUS_CLOSED;
        mqtt_conn_id = BP_API_IP_INVALID_CONNECTION_ID;
    }
}

void uart_write(const uint8_t *data, uint16_t data_len)
{
    uint16_t written = 0;

    while (written < data_len)
    {
        if (serial_port->availableForWrite() > 0)
        {
            serial_port->write(data[written]);
            written++;
        }
    }
}
