/** ********************************************************************************************************************
 * @file        MQTTClien.h
 * @author      Briowireless, Inc.
 * @date        2017-05-16
 * @brief       API layer for MQTT which simplifies the use of the API.             
 *              
 *              
 * @version     API V1
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/
#ifndef __MQTTClient__
#define __MQTTClient__

// Include directives  -------------------------------------------------------------------------------------------------

#include <bp_api.h>
#include <arduino.h>

// Preprocessor definitions --------------------------------------------------------------------------------------------

#define BP_API_MQTT_PUBLISH__CANNOT_PUBLISH_NOT_READY       0x0B    ///< Cannot publish: Underlying TCP connection is not ready.

#define BP_API_MQTT_SECURE_DEFAULT                          0
#define BP_API_MQTT_CLEAN_SESSION_DEFAULT                   1
#define BP_API_MQTT_KEEP_ALIVE_DEFAULT                      60

#define BP_API_MQTT_QOS_DEFAULT                             0
#define BP_API_MQTT_RETAINED_DEFAULT                        0

#define MAX_PAYLOAD_SIZE                                    260

#define MQTT_CONN_STATUS_CLOSED                             0
#define MQTT_CONN_STATUS_OPEN                               1
#define MQTT_CONN_STATUS_ESTABLISHED                        2

#define REQUEST_TIMEOUT                     (30*1000UL)       // 30 seconds
#define PUBLISH_TIMEOUT                     (10*1000UL)       // 10 seconds

// Global Type definitions ---------------------------------------------------------------------------------------------

typedef void (*bp_api_mqtt_init_connection_reply)(uint8_t seq_nr, uint8_t status, uint8_t mqtt_conn_id);
typedef void (*bp_api_mqtt_connection_established_notify_post)(uint8_t seq_nr, uint8_t mqtt_conn_id);
typedef void (*bp_api_mqtt_subscribe_topics_reply)(uint8_t seq_nr, uint8_t status, uint8_t sub_status);
typedef void (*bp_api_mqtt_unsubscribe_topics_reply)(uint8_t seq_nr, uint8_t status);
typedef void (*bp_api_mqtt_publish_reply)(uint8_t seq_nr, uint8_t status);
typedef void (*bp_api_mqtt_publication_rx_notify_request)(uint8_t seq_nr, uint8_t mqtt_conn_id,
                                                          uint8_t qos, uint8_t retained, uint8_t complete,
                                                          uint8_t topic_len, uint8_t* topic,
                                                          uint8_t data_len, uint8_t* data);
typedef void (*bp_api_mqtt_disconnect_reply)(uint8_t seq_nr, uint8_t status);
typedef void (*bp_api_mqtt_disconnect_notify_post)(uint8_t seq_nr, uint8_t mqtt_conn_id, uint8_t reason);

typedef void (*sub_recall_fct)(uint8_t* topic, uint8_t topic_len,
                              uint8_t* data, uint8_t data_len,
                              uint8_t qos, uint8_t retained,
                              uint8_t complete);

typedef enum 
{
    POWER_UP                 ,   //= 0,
    SET_MAX_PAYLOAD_SIZE     ,   //= 1,
    GET_HW_VERSION           ,   //= 2,
    GET_FW_VERSION           ,   //= 3,
    POWER_ON_RADIO           ,   //= 4,
    GET_RADIO_STATUS         ,   //= 5,       
    WAIT_RADIO_READY         ,   //= 6,
    SET_SIM_ACCESS           ,   //= 7,
    SET_APN_CONFIG           ,   //= 8,
    GET_MOBILE_CONN_STATUS   ,   //= 9,
    WAIT_NETWORK_REGISTRATION,   //= 10,
    MQTT_WAIT_USER_CONNECT   ,   //= 11,
    MQTT_CONN_INIT           ,   //= 12,
    MQTT_WAIT_CONNECT        ,   //= 13,
    MQTT_CONNECTED           ,   //= 14,
    MQTT_SUBSCRIBE_TOPICS    ,   //= 15,
    MQTT_UNSUBSCRIBE_TOPICS  ,   //= 16,
    MQTT_PUBLISH             ,   //= 17,
    MQTT_DISCONNECT          ,   //= 18,
    MQTT_WAIT_DISCONNECT     ,   //= 19,
    NO_SIM_DETECTED              //= 20
}State;

class MQTTClient
{
private:
  bp_api_init_params init_params;

  //Builder
  uint8_t* client_id;

  uint8_t* apn_config_host;
  uint8_t* apn_config_user;
  uint8_t* apn_config_pass;

  uint8_t* sim_pin;
  uint8_t* sim_puk;

  //Connect
  bool user_connect;

  uint8_t* host;
  uint16_t port;

  uint8_t* user;
  uint8_t* pass;

  uint8_t clean_session;
  uint8_t secure;
  uint8_t ka_ival_5_secs;

  //Subscribe
  uint8_t* subscribe_topic_1;
  uint8_t* subscribe_topic_2;
  uint8_t* subscribe_topic_3;

  uint8_t subscribe_qos_1;
  uint8_t subscribe_qos_2;
  uint8_t subscribe_qos_3;

  //Publish

  uint8_t* publish_topic;
  uint8_t* publish_data;

  uint8_t publish_qos;
  uint8_t publish_retained;

public:
  MQTTClient(uint8_t *client_id, HardwareSerial *serial_port_lc, uint8_t *apn_config_host);         

  MQTTClient(uint8_t *client_id, HardwareSerial *serial_port, uint8_t *apn_config_host,
              uint8_t *apn_config_user, uint8_t *apn_config_pass,
              uint8_t *sim_pin, uint8_t *sim_puk);

  void setup(long serial_speed, uint8_t serial_config , void (*f)(uint8_t* topic, uint8_t topic_len,
                                                                      uint8_t* data, uint8_t data_len,
                                                                      uint8_t qos, uint8_t retained,
                                                                      uint8_t complete)
                                                                    );

  void connect(uint8_t *host, uint16_t port);

  void connect(uint8_t *host, uint16_t port, 
                   uint8_t clean_session, uint8_t secure, uint8_t ka_ival_5_secs);

  void connect(uint8_t *host, uint16_t port, 
                  uint8_t *user, uint8_t *pass, uint8_t clean_session, 
                  uint8_t secure, uint8_t ka_ival_5_secs);

  void disconnect();

  void subscribe(uint8_t *topic);

  void subscribe(uint8_t *topic, uint8_t qos);

  void subscribe(uint8_t *topic1, uint8_t qos1, 
                  uint8_t *topic2, uint8_t qos2, 
                  uint8_t *topic3, uint8_t qos3);

  void unsubscribe(uint8_t *topic);

  void unsubscribe(uint8_t *topic1, uint8_t *topic2, uint8_t *topic3);

  void publish(uint8_t *topic, uint8_t *data);

  void publish(uint8_t *topic, uint8_t *data, uint8_t qos, uint8_t retained);

  bool isconnected();

  void loop();
};

// Global Variables ----------------------------------------------------------------------------------------------------

extern HardwareSerial *serial_port;

extern State state;
extern uint8_t status;
extern uint8_t mqtt_conn_id;
extern uint8_t awaiting_reply;
extern uint8_t radio_ready;
extern uint8_t mqtt_conn_status;
extern bool publish_in_progress;
extern bool network_registered;

extern unsigned long request_ts;
extern unsigned long status_request_ts;
extern unsigned long publish_ts;
extern unsigned long power_up_ts;

extern uint16_t rx_data_len;
extern uint8_t uart_rx_data;
extern char tmp_data_buffer[5];

extern uint8_t sim_status;

extern mp_seq_nr seq_number;

extern uint8_t deser_buf[MAX_PAYLOAD_SIZE + MESSAGE_PROTOCOL_HEADER_SIZE + MESSAGE_PROTOCOL_FOOTER_SIZE];
extern uint8_t ser_buf[MAX_PAYLOAD_SIZE];

// Callbacks
void sys_set_max_payload_sz_reply(uint8_t seq_nr, uint8_t status);
void devi_get_hw_version_reply(uint8_t seq_nr, uint8_t status, uint8_t hw_ver_len, uint8_t* hw_ver);
void devi_get_fw_version_reply(uint8_t seq_nr, uint8_t status, uint16_t boot_ver, uint8_t api_ver,
                               uint8_t bp_ver_len, uint8_t* bp_ver, uint8_t radio_ver_len, uint8_t* radio_ver);
void mobi_power_on_radio_reply(uint8_t seq_nr, uint8_t status);
void mobi_radio_start_notify_post(uint8_t seq_nr, uint8_t notify_reason, uint8_t sim_sta);
void mobi_get_radio_status_reply(uint8_t seq_nr, uint8_t status, uint8_t radio_status, uint8_t sim_sta);
void mobi_set_sim_access_config_reply(uint8_t seq_nr, uint8_t status);
void mobi_set_apn_config_reply(uint8_t seq_nr, uint8_t status);
void mobi_get_mobile_conn_status_reply(uint8_t seq_nr, uint8_t status, uint8_t conn_ind, uint8_t link_tech,
                                              uint16_t net_lac, uint32_t net_cell_id, uint8_t signal_strength,
                                              uint8_t* net_mcc, uint8_t* net_mnc);
void mobi_mobile_conn_status_change_notify_post(uint8_t seq_nr, uint8_t conn_ind, uint8_t link_tech,
                                                       uint16_t net_lac, uint32_t net_cell_id);
void mqtt_init_connection_reply(uint8_t seq_nr, uint8_t status, uint8_t conn_id);
void mqtt_connection_established_notify_post(uint8_t seq_nr, uint8_t conn_id);
void mqtt_subscribe_topics_reply(uint8_t seq_nr, uint8_t status, uint8_t sub_status);
void mqtt_unsubscribe_topics_reply(uint8_t seq_nr, uint8_t status);
void mqtt_publish_reply(uint8_t seq_nr, uint8_t status);
void mqtt_publication_rx_notify_request(uint8_t seq_nr, uint8_t conn_id, uint8_t qos, uint8_t retained,
                                               uint8_t complete, uint8_t topic_len, uint8_t* topic,
                                               uint8_t data_len, uint8_t* data);
void mqtt_disconnect_reply(uint8_t seq_nr, uint8_t status);
void mqtt_disconnect_notify_post(uint8_t seq_nr, uint8_t conn_id, uint8_t reason);

void uart_write(const uint8_t* data, uint16_t data_len);

void receive_data_action(uint8_t* topic, uint8_t topic_len,
                         uint8_t* data, uint8_t data_len,
                         uint8_t qos, uint8_t retained,
                         uint8_t complete);

extern bp_api_callbacks callbacks;

#endif // __MQTTClient__
