/** ********************************************************************************************************************
 * @file        SMS.ino
 * @author      Briowireless, Inc.
 * @date        2017-03-10
 * @brief       Main program file for SMS Sample.
 * @version     API V1
 * @copyright   Briowireless, Inc., 2017, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/

// Include directives  -------------------------------------------------------------------------------------------------

#include <bp_api.h>

// Preprocessor definitions --------------------------------------------------------------------------------------------

#define LED_PIN                             13

#define MAX_PAYLOAD_SIZE                    260

#ifdef ARDUINO_SAM_ZERO
#define SERIAL_PORT                         Serial1             // Serial class instance on which the BitPipe is found.
#else
#define SERIAL_PORT                         Serial              // Serial class instance on which the BitPipe is found.
#endif
#if (F_CPU > 8000000L)
#define SERIAL_SPEED                        115200
#else
#define SERIAL_SPEED                        9600
#endif
#define SERIAL_CONFIG                       SERIAL_8N1

#define REQUEST_TIMEOUT                     (30*1000UL)         // 30 seconds

#define SIM_PIN                             ""                  // When SIM card is locked using a PIN
#define SIM_PUK                             ""                  // When too many bad PINs were tried, the PIN Unlock Key is necessary.

#define APN_CONFIG_HOST                     ""                  // TODO: Enter you APN host address here
#define APN_CONFIG_USER                     ""
#define APN_CONFIG_PASS                     ""

#define SMS_RECIPIENT_MSISDN                "15555555555"       // SMS recipient mobile phone number (In Canada and US the 11-digit phone number). Do not prefix with '+'.
#define SMS_ENCODING                        0                   // 7-bit GSM 03.38

#define SMS_TMP_BLOCK_LEN                   32
#define SMS_MESSAGE_LEN                     107

typedef enum 
{
    POWER_UP,
    SET_MAX_PAYLOAD_SIZE,
    GET_HW_VERSION,
    GET_FW_VERSION,
    POWER_ON_RADIO,
    GET_RADIO_STATUS,
    WAIT_RADIO_READY,
    SET_SIM_ACCESS,
    SET_APN_CONFIG,
    GET_MOBILE_CONN_STATUS,
    WAIT_NETWORK_REGISTRATION,
    SMS_SETUP,
    SMS_WRITE,
    SMS_SEND,
    DONE,
    NO_SIM_DETECTED,
}State;

// Local functions prototypes ------------------------------------------------------------------------------------------

// Callbacks
void sys_set_max_payload_sz_reply(uint8_t seq_nr, uint8_t status);
void devi_get_hw_version_reply(uint8_t seq_nr, uint8_t status, uint8_t hw_ver_len, uint8_t* hw_ver);
void devi_get_fw_version_reply(uint8_t seq_nr, uint8_t status, uint16_t boot_ver, uint8_t api_ver,
                               uint8_t bp_ver_len, uint8_t* bp_ver, uint8_t radio_ver_len, uint8_t* radio_ver);
void mobi_power_on_radio_reply(uint8_t seq_nr, uint8_t status);
void mobi_radio_start_notify_post(uint8_t seq_nr, uint8_t notify_reason, uint8_t sim_sta);
void mobi_get_radio_status_reply(uint8_t seq_nr, uint8_t status, uint8_t radio_status, uint8_t sim_sta);
void mobi_set_sim_access_config_reply(uint8_t seq_nr, uint8_t status);
void mobi_set_apn_config_reply(uint8_t seq_nr, uint8_t status);
void mobi_get_mobile_conn_status_reply(uint8_t seq_nr, uint8_t status, uint8_t conn_ind, uint8_t link_tech,
                                              uint16_t net_lac, uint32_t net_cell_id, uint8_t signal_strength,
                                              uint8_t* net_mcc, uint8_t* net_mnc);
void mobi_mobile_conn_status_change_notify_post(uint8_t seq_nr, uint8_t conn_ind, uint8_t link_tech,
                                                       uint16_t net_lac, uint32_t net_cell_id);
void sms_setup_transmission_reply(uint8_t seq_nr, uint8_t status);
void sms_message_append_send_reply(uint8_t seq_nr, uint8_t status);
void sms_message_rx_notify_request(uint8_t seq_nr, uint8_t encoding, uint8_t year, uint8_t month, uint8_t day,
                                            uint8_t hour, uint8_t minute, uint8_t second,
                                            uint8_t tz_info_offset, uint8_t tz_info_sign,
                                            uint8_t msisdn_len, uint8_t* msisdn);
void sms_message_part_notify_request(uint8_t seq_nr, uint8_t last_part_indicator, uint8_t data_len, uint8_t* data);

void uart_write(const uint8_t* data, uint16_t data_len);

// Local variables  ----------------------------------------------------------------------------------------------------

// Buffers used by SDK.
uint8_t deser_buf[MAX_PAYLOAD_SIZE + MESSAGE_PROTOCOL_HEADER_SIZE + MESSAGE_PROTOCOL_FOOTER_SIZE];
uint8_t ser_buf[MAX_PAYLOAD_SIZE];

bp_api_init_params init_params;

State state;
mp_seq_nr seq_number;
bool awaiting_reply;
unsigned long request_ts;
bool radio_ready;
uint8_t sim_status;
bool network_registered;

unsigned long status_request_ts;

char sms_tmp_buffer[SMS_TMP_BLOCK_LEN];
uint8_t sms_bytes_written;
size_t sms_bytes_sent;

#ifdef ARDUINO_ARCH_AVR
const char sms_message[SMS_MESSAGE_LEN+1] PROGMEM =
#else
const char sms_message[SMS_MESSAGE_LEN+1] =
#endif
{
    "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Duis accumsan, ante vitae cursus venenatis, lacus."
};

/* Unused callbacks can be set to NULL */
bp_api_callbacks callbacks  = {
    NULL,                                       // bp_api_sys_error_report_post
    sys_set_max_payload_sz_reply,               // bp_api_sys_set_max_payload_sz_reply
    NULL,                                       // bp_api_sys_set_uart_speed_reply

    devi_get_hw_version_reply,                  // bp_api_devi_get_hw_version_reply
    devi_get_fw_version_reply,                  // bp_api_devi_get_fw_version_reply
    NULL,                                       // bp_api_devi_get_product_name_reply
    NULL,                                       // bp_api_devi_get_vin_vio_voltage_reply
    NULL,                                       // bp_api_devi_get_junction_temp_reply
    NULL,                                       // bp_api_devi_set_low_power_reply
    NULL,                                       // bp_api_devi_get_latest_fw_version_reply
    NULL,                                       // bp_api_devi_update_fw_reply
    NULL,                                       // bp_api_devi_get_autonomous_mode_reply
    NULL,                                       // bp_api_devi_set_autonomous_mode_reply

    mobi_power_on_radio_reply,                  // bp_api_mobi_power_on_radio_reply
    NULL,                                       // bp_api_mobi_power_off_radio_reply
    mobi_radio_start_notify_post,               // bp_api_mobi_radio_start_notify_post
    mobi_get_radio_status_reply,                // bp_api_mobi_get_radio_status_reply
    NULL,                                       // bp_api_mobi_get_imei_reply
    NULL,                                       // bp_api_mobi_get_iccid_reply
    NULL,                                       // bp_api_mobi_get_imsi_reply
    NULL,                                       // bp_api_mobi_get_own_phone_number_reply
    NULL,                                       // bp_api_mobi_get_apn_config_reply
    mobi_set_apn_config_reply,                  // bp_api_mobi_set_apn_config_reply
    mobi_set_sim_access_config_reply,           // bp_api_mobi_set_sim_access_config_reply
    mobi_get_mobile_conn_status_reply,          // bp_api_mobi_get_mobile_conn_status_reply
    mobi_mobile_conn_status_change_notify_post, // bp_api_mobi_mobile_conn_status_change_notify_post
    NULL,                                       // bp_api_mobi_get_mobile_date_time_reply

    NULL,                                       // bp_api_ip_tcp_init_connection_reply
    NULL,                                       // bp_api_ip_tcp_connection_established_notify_post
    NULL,                                       // bp_api_ip_tcp_write_append_reply
    NULL,                                       // bp_api_ip_tcp_write_send_reply
    NULL,                                       // bp_api_ip_tcp_rx_notify_request
    NULL,                                       // bp_api_ip_http_init_connection_reply
    NULL,                                       // bp_api_ip_http_post_write_ready_notify_post
    NULL,                                       // bp_api_ip_http_post_write_reply
    NULL,                                       // bp_api_ip_http_rx_notify_request
    NULL,                                       // bp_api_ip_connection_close_reply
    NULL,                                       // bp_api_ip_connection_close_notify_post

    NULL,                                       // bp_api_mqtt_init_connection_reply
    NULL,                                       // bp_api_mqtt_connection_established_notify_post
    NULL,                                       // bp_api_mqtt_subscribe_topics_reply
    NULL,                                       // bp_api_mqtt_unsubscribe_topics_reply
    NULL,                                       // bp_api_mqtt_publish_reply
    NULL,                                       // bp_api_mqtt_publication_rx_notify_request
    NULL,                                       // bp_api_mqtt_disconnect_reply
    NULL,                                       // bp_api_mqtt_disconnect_notify_post

    sms_setup_transmission_reply,               // bp_api_sms_setup_transmission_reply
    sms_message_append_send_reply,              // bp_api_sms_message_append_send_reply
    sms_message_rx_notify_request,              // bp_api_sms_message_rx_notify_request
    sms_message_part_notify_request,            // bp_api_sms_message_part_notify_request
};

// Local functions  ----------------------------------------------------------------------------------------------------

void sys_set_max_payload_sz_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = GET_HW_VERSION;
        awaiting_reply = false;
    }
}

void devi_get_hw_version_reply(uint8_t seq_nr, uint8_t status, uint8_t hw_ver_len, uint8_t* hw_ver)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = GET_FW_VERSION;
        awaiting_reply = false;
    }
}

void devi_get_fw_version_reply(uint8_t seq_nr, uint8_t status, uint16_t boot_ver, uint8_t api_ver,
                               uint8_t bp_ver_len, uint8_t* bp_ver, uint8_t radio_ver_len, uint8_t* radio_ver)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = POWER_ON_RADIO;
        awaiting_reply = false;
    }
}

void mobi_power_on_radio_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        (status == MSG_STATUS_CODE_SUCCESS || status == BP_API_PWRON_RADIO__STATUS_ALREADY_STARTED))
    {
        if (status == MSG_STATUS_CODE_SUCCESS)
        {
            radio_ready = false;
            state = GET_RADIO_STATUS;
        }
        else if (status == BP_API_PWRON_RADIO__STATUS_ALREADY_STARTED)
        {
            radio_ready = true;
            state = SET_SIM_ACCESS;
        }
        awaiting_reply = false;
    }
}

void mobi_radio_start_notify_post(uint8_t seq_nr, uint8_t notify_reason, uint8_t sim_sta)
{
    // Keep track if we're registered on the network.
    radio_ready = (BP_API_RADIO_START_NOTIFY__REASON_PWRON_READY == notify_reason ||
                       BP_API_RADIO_START_NOTIFY__REASON_UNEXPECTED_REBOOT == notify_reason);

    sim_status = sim_sta;
    awaiting_reply = false;
}

void mobi_get_radio_status_reply(uint8_t seq_nr, uint8_t status, uint8_t radio_status, uint8_t sim_sta)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        radio_ready = (BP_API_GET_RADIO_STATUS__RADIO_PWRON_READY == radio_status);

        sim_status = sim_sta;

        state = WAIT_RADIO_READY;
        status_request_ts = millis();
        awaiting_reply = false;
    }
}

void mobi_set_sim_access_config_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = SET_APN_CONFIG;
        awaiting_reply = false;
    }
}

void mobi_set_apn_config_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = GET_MOBILE_CONN_STATUS;
        awaiting_reply = false;
    }
}

void mobi_get_mobile_conn_status_reply(uint8_t seq_nr, uint8_t status, uint8_t conn_ind, uint8_t link_tech,
                                              uint16_t net_lac, uint32_t net_cell_id, uint8_t signal_strength,
                                              uint8_t* net_mcc, uint8_t* net_mnc)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        network_registered = (BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_REGD_HOME == conn_ind ||
                              BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_REGD_FOREIGN == conn_ind);
        state = WAIT_NETWORK_REGISTRATION;
        status_request_ts = millis();
        awaiting_reply = false;
    }
}

void mobi_mobile_conn_status_change_notify_post(uint8_t seq_nr, uint8_t conn_ind, uint8_t link_tech,
                                                       uint16_t net_lac, uint32_t net_cell_id)
{
    if (!network_registered)
    {
        // If we are not connected and just received a status change for mobile connection, re-ask mobile connection status to BitPipe.
        state = GET_MOBILE_CONN_STATUS;
    }
    else if (BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_NOT_CONNECTED == conn_ind ||
             BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_CAMPING == conn_ind)
    {
        network_registered = false;
    }
}

void sms_setup_transmission_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        state = SMS_WRITE;
        awaiting_reply = false;
    }
}

void sms_message_append_send_reply(uint8_t seq_nr, uint8_t status)
{
    if (seq_nr == seq_number &&
        status == MSG_STATUS_CODE_SUCCESS)
    {
        if (state == SMS_WRITE)
        {
            sms_bytes_sent += sms_bytes_written;
        }
        else
        {
            state = DONE;
        }
        awaiting_reply = false;
    }
}

void sms_message_rx_notify_request(uint8_t seq_nr, uint8_t encoding, uint8_t year, uint8_t month, uint8_t day,
                                          uint8_t hour, uint8_t minute, uint8_t second,
                                          uint8_t tz_info_offset, uint8_t tz_info_sign,
                                          uint8_t msisdn_len, uint8_t* msisdn)
{
    // IMPORTANT: Reply to this request, or else the BitPipe will re-send this notification.
    bp_api_sms_message_rx_notify_ack(seq_nr);

    // Do anything intelligent with the message metadata
}

void sms_message_part_notify_request(uint8_t seq_nr, uint8_t last_part_indicator, uint8_t data_len, uint8_t* data)
{
    // IMPORTANT: Reply to this request, or else the BitPipe will re-send this notification.
    bp_api_sms_message_part_notify_ack(seq_nr);

    // Do anything intelligent with the message received
}

void uart_write(const uint8_t* data, uint16_t data_len)
{
    uint16_t written = 0;

    while(written < data_len)
    {
        if (SERIAL_PORT.availableForWrite() > 0)
        {
            SERIAL_PORT.write(data[written]);
            written++;
        }
    }
}

void setup()
{
    uint8_t status;                                     // Request execution status.
    int i;

    digitalWrite(LED_PIN, LOW);
    pinMode(LED_PIN, OUTPUT);

    // Initialize Serial
    SERIAL_PORT.begin(SERIAL_SPEED, SERIAL_CONFIG);

    // Initialize SDK layer.
    init_params.app_callbacks       =   &callbacks;
    init_params.uart_write =            &uart_write;
    init_params.deserialize_buf     =   deser_buf;
    init_params.deserialize_buf_len =   sizeof(deser_buf);
    init_params.serialize_buf       =   ser_buf;
    init_params.serialize_buf_len   =   sizeof(ser_buf);

    status = bp_api_init(&init_params);
    if(status != 0)
    {
        digitalWrite(LED_PIN, HIGH);
        while(1);
    }

    awaiting_reply = false;
    radio_ready = false;
    network_registered = false;
    state = POWER_UP;
}

void loop()
{
    uint8_t status;                                 // Request execution status.
    uint16_t rx_data_len;
    uint8_t uart_rx_data;
    uint8_t led_state;
    unsigned long ts;
    static unsigned long power_up_ts = millis();

    if (state < DONE)
    {
        digitalWrite(LED_PIN, millis() % 1000 < 50 ? HIGH : LOW);
    }

    // Check if we've received data from the BitPipe, and process it.
    if (SERIAL_PORT.available() > 0)
    {
        uart_rx_data = (uint8_t)SERIAL_PORT.read();
        bp_api_process_incoming(&uart_rx_data, sizeof(uart_rx_data), &rx_data_len);
    }

    if(!awaiting_reply)
    {
        status = 0xFF;
        switch(state)
        {
            case POWER_UP:
                if ((millis() - power_up_ts) > 5000UL)
                {
                    state = SET_MAX_PAYLOAD_SIZE;
                }
                break;

            case SET_MAX_PAYLOAD_SIZE:
                // Configure Maximum message payload size.
                status = bp_api_sys_set_max_payload_sz(&seq_number, MAX_PAYLOAD_SIZE);
                // state is advance when receive reply from bitpipe
                break;

            case GET_HW_VERSION:
                // Get HW version.
                status = bp_api_devi_get_hw_version(&seq_number);
                break;

            case GET_FW_VERSION:
                // Get FW version.
                status = bp_api_devi_get_fw_version(&seq_number);
                break;

            case POWER_ON_RADIO:
                // Power-on Radio.
                status = bp_api_mobi_power_on_radio(&seq_number,
                                                    BP_API_PWRON_RADIO__PWRON_METHOD_DCDC,
                                                    BP_API_PWRON_RADIO__BACKOFF_DISABLED);
                break;

            case GET_RADIO_STATUS:
                // Get Radio Status.
                status = bp_api_mobi_get_radio_status(&seq_number);
                break;

            case WAIT_RADIO_READY:
                // Wait for radio to be ready.
                if(radio_ready)
                {
                    state = SET_SIM_ACCESS;
                }
                else if ((millis() - status_request_ts) > REQUEST_TIMEOUT)
                {
                    state = GET_RADIO_STATUS;
                }
                break;

            case SET_SIM_ACCESS:     // Check SIM card status.
                if(BP_API_RADIO_START_NOTIFY__SIM_NOT_INSERTED == sim_status)
                {
                    state = NO_SIM_DETECTED;
                }
                else if(BP_API_RADIO_START_NOTIFY__SIM_DENIED_NEED_PIN == sim_status)
                {
                    status = bp_api_mobi_set_sim_access_config(&seq_number, strlen(SIM_PIN), (uint8_t*)SIM_PIN);
                }
                else if(BP_API_RADIO_START_NOTIFY__SIM_DENIED_NEED_PUK == sim_status)
                {
                    status = bp_api_mobi_set_sim_access_config(&seq_number, strlen(SIM_PUK), (uint8_t*)SIM_PUK);
                }
                else    // SIM is OK.
                {
                    state = SET_APN_CONFIG;
                }
                break;

            case SET_APN_CONFIG:
                // Set APN Configuration (This doesn't need to be set every time).
                status = bp_api_mobi_set_apn_config(&seq_number,
                                                    strlen(APN_CONFIG_HOST), (uint8_t*)APN_CONFIG_HOST,
                                                    strlen(APN_CONFIG_USER), (uint8_t*)APN_CONFIG_USER,
                                                    strlen(APN_CONFIG_PASS), (uint8_t*)APN_CONFIG_PASS);
                break;

            case GET_MOBILE_CONN_STATUS:
                // Get Mobile connection status.
                status = bp_api_mobi_get_mobile_conn_status(&seq_number);
                break;

            case WAIT_NETWORK_REGISTRATION:
                // Wait for network registration.
                if(network_registered)
                {
                    state = SMS_SETUP;
                }
                // If no message received in 30 seconds, re-ask for status
                else if ((millis() - status_request_ts) > REQUEST_TIMEOUT)
                {
                    // This will trigger a re-send of the get mobile conn status (with a different sequence number).
                    state = GET_MOBILE_CONN_STATUS;
                }
                break;

            case SMS_SETUP:
                // SMS: Setup transmission
                status = bp_api_sms_setup_transmission(&seq_number, 0, SMS_ENCODING, strlen(SMS_RECIPIENT_MSISDN), (uint8_t*)SMS_RECIPIENT_MSISDN);
                break;

            case SMS_WRITE:
                if (sms_bytes_sent < SMS_MESSAGE_LEN)
                {
                    sms_bytes_written = (SMS_MESSAGE_LEN - sms_bytes_sent) > SMS_TMP_BLOCK_LEN ?
                                            SMS_TMP_BLOCK_LEN :
                                            (SMS_MESSAGE_LEN - sms_bytes_sent);
#ifdef ARDUINO_ARCH_AVR
                    for (int i = 0; i < sms_bytes_written; i++)
                    {
                        sms_tmp_buffer[i] = pgm_read_byte(&sms_message[sms_bytes_sent]);
                    }
#else
                    memcpy(sms_tmp_buffer, &sms_message[sms_bytes_sent], sms_bytes_written);
#endif
                    // Write SMS message (in BitPipe internal buffer)
                    status = bp_api_sms_message_append_send(&seq_number, false, sms_bytes_written, (uint8_t*)sms_tmp_buffer);
                }
                else
                {
                    state = SMS_SEND;
                }
                break;

            case SMS_SEND:
                // Send the SMS (from BitPipe internal buffer) to the recipient.
                status = bp_api_sms_message_append_send(&seq_number, true, 0, NULL);
                break;

            case DONE:
                led_state = HIGH;
                while (1)
                {
                    digitalWrite(LED_PIN, led_state);
                    led_state = led_state == HIGH ? LOW : HIGH;

                    ts = millis();
                    while ((millis() - ts) < 500)
                    {
                        // Check if we've received data from the BitPipe, and process it (to receive SMS).
                        if (SERIAL_PORT.available() > 0)
                        {
                            uart_rx_data = (uint8_t)SERIAL_PORT.read();
                            bp_api_process_incoming(&uart_rx_data, sizeof(uart_rx_data), &rx_data_len);
                        }
                    }
                }
                break;

            case NO_SIM_DETECTED:
                // No SIM is inserted, we can do nothing
                while (1)
                {
                    digitalWrite(LED_PIN, HIGH);
                    delay(100);
                    digitalWrite(LED_PIN, LOW);
                    delay(100);
                }
                break;
        }

        if (status == 0)
        {
            awaiting_reply = true;
            request_ts = millis();
        }
    }
    else
    {
        // Check for sent request timeout.
        if ((millis() - request_ts) > REQUEST_TIMEOUT)
        {
            // This will trigger a re-send of the request (with a different sequence number).
            awaiting_reply = false;
        }
    }
}

