/** ********************************************************************************************************************
 * @file        bp_api_devi.c
 * @author      Briowireless, Inc.
 * @date        2016-10-28
 * @brief       API layer for DEVICE INFORMATION requests to the BitPipe.
 *              See the BitPipe Serial API document for more information on requests below.
 * @version     API V1
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/

// Include directives  -------------------------------------------------------------------------------------------------

#include <bp_api_devi.h>
#include <bp_api.h>
#include <message_defs.h>

#ifdef __cplusplus
extern "C" {
#endif

// Preprocessor definitions --------------------------------------------------------------------------------------------

// Global (extern) function  -------------------------------------------------------------------------------------------

/**
 * @defgroup api_devi BitPipe API: Device Information
 * Allows querying and configuring values regarding the BitPipe device, not related to wireless connectivity.
 * @{
 */

/** ********************************************************************************************************************
 * @brief  Retrieves the hardware version of the BitPipe.
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_devi_get_hw_version(mp_seq_nr* seq_nr)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Querying HW ver ...\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_GET_HW_VERSION, msg_seq_nr, 0x00, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Retrieves the firmware version of the BitPipe.
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_devi_get_fw_version(mp_seq_nr* seq_nr)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Querying FW ver ...\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_GET_FW_VERSION, msg_seq_nr, 0x00, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Retrieves the Product Name of the BitPipe.
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_devi_get_product_name(mp_seq_nr* seq_nr)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Querying Product Name ...\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_GET_PRODUCT_NAME, msg_seq_nr, 0x00, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Retrieves the voltage values of VIN (Wireless slave input voltage) and VIO (BitPipe internal voltage).
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_devi_get_vin_vio_voltage(mp_seq_nr* seq_nr)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Querying VIO Voltage ...\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_GET_VIN_VIO_VOLTAGE, msg_seq_nr, 0x00, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Retrieves the internal temperature of the BitPipe.
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_devi_get_junction_temp(mp_seq_nr* seq_nr)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Querying Junction Temperature ...\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_GET_JUNCTION_TEMP, msg_seq_nr, 0x00, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Puts the BitPipe in low power consumption mode.
 *         During this time, the BitPipe is put into a deep sleep mode, and can only be wakened by sending it a
 *         single byte on the UART port.
 *
 * @note   If you send a message to the BitPipe while it is in low power, the message will be dropped while the BitPipe
 *         exits low power mode. It is therefore recommended to exit low power mode by sending a single byte, not a
 *         message.
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_devi_set_low_power_mode(mp_seq_nr* seq_nr)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Set Low Power Mode ...\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_SET_LOW_POWER_MODE, msg_seq_nr, 0x00, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Retrieves the latest firmware version of the BitPipe.
 *         This internally triggers an HTTP connection to Briowireless's servers to check the latest version for this
 *         BitPipe product flavor.
 *         This request may take some time to complete, depending on network quality.
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_devi_get_latest_fw_version(mp_seq_nr* seq_nr)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Querying Latest FW Version ...\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_GET_LATEST_FW_VERSION, msg_seq_nr, 0x00, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief Instructs the BitPipe to update its firmware version to the target version specified in parameter 'target'.
 *        The BitPipe will initiate an HTTP connection to Briowireless's server to validate that the target firmware
 *        version exists before jumping in the Bootloader and performing the update.
 * @note  Use the target version 'latest' to update to the latest stable release.
 *
 * @param  seq_nr       The message sequence number of the request sent to the BitPipe, if successful.
 * @param  target       The firmware version to update to. Use 'latest' for updating to the latest stable release.
 * @param  target_len   The length of the target firmware version (excluding NULL string terminator, if any).
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_devi_update_fw(mp_seq_nr* seq_nr, uint8_t* target, uint8_t target_len)
{
    uint8_t linear_buf[1+target_len];
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    linear_buf[0] = target_len;
    memcpy(&linear_buf[1], target, target_len);

    P_WARN("[%02X] Update Firmware ...\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_UPDATE_FIRMWARE, msg_seq_nr, 0x00, linear_buf, sizeof(linear_buf));
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Retrieves the configuration of the AUTONOMOUS MODE.
 *         The autonomous mode configuration is basically a set of MQTT parameters the BitPipe uses to automatically
 *         connect to, when the autonomous/modem switch is set to 'autonomous'.
 * @see    The BitPipe dev-kit userguide for more information on AUTONOMOUS MODE.
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_devi_get_autonomous_mode_config(mp_seq_nr* seq_nr)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Get Autonomous Mode Configuration ...\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_GET_AUTONOMOUS_CONFIG, msg_seq_nr, 0x00, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Sets the configuration of the AUTONOMOUS MODE.
 *         The autonomous mode configuration is basically a set of MQTT parameters the BitPipe uses to automatically
 *         connect to, when the autonomous/modem switch is set to 'autonomous'.
 * @see    The BitPipe dev-kit userguide for more information on AUTONOMOUS MODE.
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 * @param  secure   Indicates whether to use SSL/TLS. Values: [0,1]
 * @param  ka_ival  Keep-alive interval at the MQTT layer, in seconds. Values: [0,300-1200]
 * @param  port     MQTT TCP port of the server.
 * @param  host_len Length of MQTT host (or IP address) of the server (excluding NULL string terminator, if any).
 * @param  host     MQTT host (or IP address) of the server.
 * @param  user_len Length of the authentication user name for MQTT CONNECT (excluding NULL string terminator, if any).
 * @param  user     Authentication user name for MQTT CONNECT.
 * @param  pass_len Length of the authentication password during MQTT CONNECT (excluding NULL string terminator, if any).
 * @param  pass     Authentication password during MQTT CONNECT.
 * @param  gpfx_len Length of the global prefix (excluding NULL string terminator, if any).
 * @param  gpfx     Global prefix prepended to all topics used to subscribe, unsubscribe or publish.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_devi_set_autonomous_mode_config(mp_seq_nr* seq_nr, uint8_t secure, uint16_t ka_ival, uint16_t port,
                                               uint8_t host_len, uint8_t* host, uint8_t user_len, uint8_t* user,
                                               uint8_t pass_len, uint8_t* pass, uint8_t gpfx_len, uint8_t* gpfx)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;
    uint8_t linear_buf[9+host_len+user_len+pass_len+gpfx_len];

    linear_buf[0] = secure;
    linear_buf[1] = (ka_ival >> 8) & 0xFF;
    linear_buf[2] = ka_ival & 0xFF;
    linear_buf[3] = (port >> 8) & 0xFF;
    linear_buf[4] = port & 0xFF;
    linear_buf[5] = host_len;
    memcpy(&linear_buf[6], host, host_len);

    linear_buf[6+host_len] = user_len;
    memcpy(&linear_buf[7+host_len], user, user_len);
    linear_buf[7+host_len+user_len] = pass_len;
    memcpy(&linear_buf[8+host_len+user_len], pass, pass_len);
    linear_buf[8+host_len+user_len+pass_len] = gpfx_len;
    memcpy(&linear_buf[9+host_len+user_len+pass_len], gpfx, gpfx_len);

    P_WARN("[%02X] Set Autonomous Mode Configuration ...\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_SET_AUTONOMOUS_CONFIG, msg_seq_nr, 0x00, linear_buf, sizeof(linear_buf));
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

// ---------------------------------------------------------------------------------------------------------------------
// Doxygen documentation support.
// ---------------------------------------------------------------------------------------------------------------------

/**
 * @typedef void (*bp_api_devi_get_hw_version_reply)(uint8_t seq_nr, uint8_t status, uint8_t hw_ver_len, uint8_t* hw_ver)
 * @brief  Callback into application when the reply to GET HW VERSION request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  hw_ver_len   Length of the hardware version string.
 * @param  hw_ver       The hardware version (ASCII).
 */

/**
 * @typedef void (*bp_api_devi_get_fw_version_reply)(uint8_t seq_nr, uint8_t status, uint16_t boot_ver, uint8_t api_ver, uint8_t bp_ver_len, uint8_t* bp_ver, uint8_t radio_ver_len, uint8_t* radio_ver)
 * @brief  Callback into application when the reply to GET FW VERSION request is received from the BitPipe.
 *
 * @note   The radio firmware version is only reported when the radio is powered-on and ready.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  boot_ver     The BitPipe bootloader version. The high byte indicates the major version, the low byte indicates the minor version.
 * @param  api_ver      The BitPipe API version, numeric.
 * @param  bp_ver_len   The length of the BitPipe firmware version.
 * @param  bp_ver       The BitPipe firmware version (ASCII).
 * @param  radio_ver_len The length of the radio firmware version.
 * @param  radio_ver    The radio firmware version (ASCII).
 */

/**
 * @typedef void (*bp_api_devi_get_product_name_reply)(uint8_t seq_nr, uint8_t status, uint8_t prod_name_len, uint8_t* prod_name)
 * @brief  Callback into application when the reply to GET PRODUCT NAME request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  prod_name_len The length of the BitPipe product name.
 * @param  prod_name    The BitPipe product name (i.e. *Flavor*).
 */

/**
 * @typedef void (*bp_api_devi_get_vin_vio_voltage_reply)(uint8_t seq_nr, uint8_t status, uint16_t vin_voltage, uint16_t vio_voltage)
 * @brief  Callback into application when the reply to GET VIN VIO VOLTAGE request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  vin_voltage  The voltage, in millivolts, sensed at VIN.
 * @param  vio_voltage  The voltage, in millivolts, sensed at VIO.
 */

/**
 * @typedef void (*bp_api_devi_get_junction_temp_reply)(uint8_t seq_nr, uint8_t status, int8_t junction_temp)
 * @brief  Callback into application when the reply to GET JUNCTION TEMP request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  junction_temp The junction temperature, in celcius degrees. Attention: This byte is signed.
 */

/**
 * @typedef void (*bp_api_devi_set_low_power_reply)(uint8_t seq_nr, uint8_t status)
 * @brief  Callback into application when the reply to SET LOW POWER request is received from the BitPipe.
 *
 * @note   When status indicates success, it means the BitPipe is in low-power mode. To wake the BitPipe, send a single
 *         character on the UART link.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @typedef void (*bp_api_devi_get_latest_fw_version_reply)(uint8_t seq_nr, uint8_t status, uint8_t fw_ver_len, uint8_t* fw_ver)
 * @brief  Callback into application when the reply to GET LATEST FW VERSION request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  fw_ver_len   The length of the latest firmware version found on Briowireless's servers.
 * @param  fw_ver       The latest firmware version found on Briowireless's servers.
 */

/**
 * @typedef void (*bp_api_devi_update_fw_reply)(uint8_t seq_nr, uint8_t status)
 * @brief  Callback into application when the reply to UPDATE FIRMWARE request is received from the BitPipe.
 *
 * @note   When the status indicates success, the BitPipe will become offline to perform the firmware update operation.
 *         This operation takes 3-5 minutes in good network conditions. There is no notification when the update has
 *         completed. It is suggested to issue a GET FW VERSION request every 5 seconds until the BitPipe replies with
 *         an answer.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @typedef void (*bp_api_devi_get_autonomous_mode_config_reply)(uint8_t seq_nr, uint8_t status, uint8_t secure, uint16_t ka_ival, uint16_t port, uint8_t host_len, uint8_t* host, uint8_t user_len, uint8_t* user, uint8_t pass_len, uint8_t* pass, uint8_t gpfx_len, uint8_t* gpfx)
 * @brief  Callback into application when the reply to GET AUTONOMOUS MODE CONFIGURATION request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  secure       Indicates whether to use SSL/TLS. Values: [0,1]
 * @param  ka_ival      Keep-alive interval at the MQTT layer, in seconds. Values: [0,300-1200]
 * @param  port         MQTT TCP port of the server.
 * @param  host_len     Length of MQTT host (or IP address) of the server.
 * @param  host         MQTT host (or IP address) of the server.
 * @param  user_len     Length of the authentication user name for MQTT CONNECT.
 * @param  user         Authentication user name for MQTT CONNECT.
 * @param  pass_len     Length of the authentication password during MQTT CONNECT.
 * @param  pass         Authentication password during MQTT CONNECT.
 * @param  gpfx_len     Length of the global prefix.
 * @param  gpfx         Global prefix prepended to all topics used to subscribe, unsubscribe or publish.
 */

/**
 * @typedef void (*bp_api_devi_set_autonomous_mode_config_reply)(uint8_t seq_nr, uint8_t status)
 * @brief  Callback into application when the reply to SET AUTONOMOUS MODE CONFIGURATION request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @}       // Doxygen api_devi defgroup
 */

#ifdef __cplusplus
}
#endif
