/** ********************************************************************************************************************
 * @file        bp_api_ip.c
 * @author      Briowireless, Inc.
 * @date        2016-11-03
 * @brief       API layer for IP CONNECTIVITY requests and replies to the BitPipe, and defining callback prototypes for
 *               requests, replies and asynchronous posts sent from the BitPipe.
 *              See the BitPipe Serial API document for more information on requests below.
 * @version     API V1
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/

// Include directives  -------------------------------------------------------------------------------------------------

#include <bp_api_ip.h>
#include <bp_api.h>
#include <message_defs.h>

#ifdef __cplusplus
extern "C" {
#endif

// Preprocessor definitions --------------------------------------------------------------------------------------------

// Global (extern) function  -------------------------------------------------------------------------------------------

/**
 * @defgroup api_ip BitPipe API: IP Connectivity
 * Allows initiating outbund connections to TCP / HTTP server endpoints on the Internet, transferring data both ways,
 *  and handling disconnection.
 * @{
 */

/** ********************************************************************************************************************
 * @brief  Initiates a TCP client connection to a remote TCP server.
 *
 * @param  seq_nr       The message sequence number of the request sent to the BitPipe, if successful.
 * @param  secure       Specifies if the TCP connection should be encrypted using SSL/TLS. Values: [0,1]
 * @param  conn_class   TCP connection class. Values: [0-2]. For now only connection class 'A' (value=0) is defined.
 * @param  host_len     Length of the remote TCP server hostname, or IP address (excluding NULL string terminator, if any).
 * @param  host         Remote TCP server hostname, or IP address.
 * @param  port         TCP remote port to connect to.
 * @param  ka_ival      Keep-alive interval, in seconds, at the TCP layer. Not supported by all product flavors. Value: 0
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_ip_tcp_init_connection(mp_seq_nr* seq_nr, uint8_t secure, uint8_t conn_class, uint8_t host_len, uint8_t* host, uint16_t port, uint16_t ka_ival)
{
    uint8_t linear_buf[2+host_len+2+2];
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    linear_buf[0] = 0;
    if(secure)
    {
        linear_buf[0] |= (1<<0);        // SSL/TLS
    }
    linear_buf[0] |= (conn_class << 6);

    linear_buf[1] = host_len;
    memcpy(&linear_buf[2], host, host_len);

    linear_buf[2+host_len] = (port >> 8) & 0xFF;
    linear_buf[3+host_len] = port & 0xFF;

    linear_buf[4+host_len] = (ka_ival >> 8) & 0xFF;
    linear_buf[5+host_len] = ka_ival & 0xFF;

    P_WARN("[%02X] Initiating %sTCP connection to %s:%d ...\n", msg_seq_nr, secure ? "secure ":"", host, port);

    uint8_t status = bp_api_send_message(MSG_ID_TCP_INITIATE_CONNECTION, msg_seq_nr, 0x00, linear_buf, sizeof(linear_buf));
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Appends TCP user data in the BitPipe internal buffer, for an eventual send.
 *         This operation allows you to append data in an internal buffer in the BitPipe to maximize data payload versus
 *         Ethernet/IP/TCP headers.
 *         When buffer is full, or all data to be written is in the BitPipe, use the bp_api_ip_tcp_write_send() function
 *         to actually transmit it on the Internet.
 *
 * @param  seq_nr       The message sequence number of the request sent to the BitPipe, if successful.
 * @param  tcp_conn_id  A valid TCP connection identifier acquired in the reply to the TCP INITIATE CONNECTION request.
 * @param  data_len     Length of the user data in this message.
 * @param  data         User data.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_ip_tcp_write_append(mp_seq_nr* seq_nr, uint8_t tcp_conn_id, uint8_t data_len, uint8_t* data)
{
    uint8_t linear_buf[2+data_len];
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Writing data on TCP connection %d ...\n", msg_seq_nr, tcp_conn_id);

    linear_buf[0] = tcp_conn_id;
    linear_buf[1] = data_len;
    memcpy(&linear_buf[2], data, data_len);

    uint8_t status = bp_api_send_message(MSG_ID_TCP_WRITE_APPEND, msg_seq_nr, 0x00, linear_buf, sizeof(linear_buf));
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Sends the user TCP data accumulated internally in the BitPipe to the remote TCP endpoint.
 *
 * @param  seq_nr       The message sequence number of the request sent to the BitPipe, if successful.
 * @param  tcp_conn_id  A valid TCP connection identifier acquired in the reply to the TCP INITIATE CONNECTION request.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_ip_tcp_write_send(mp_seq_nr* seq_nr, uint8_t tcp_conn_id)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;
    P_WARN("[%02X] Sending data on TCP connection %d ...\n", msg_seq_nr, tcp_conn_id);

    uint8_t status = bp_api_send_message(MSG_ID_TCP_WRITE_SEND, msg_seq_nr, 0x00, &tcp_conn_id, 1);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Sends a reply (as acknowledgement) to the BitPipe, indicating that the TCP RX NOTIFY request was well
 *         received by the Host.
 *
 * @note   Failure to acknowledge TCP RX NOTIFY requests from the BitPipe will result in the BitPipe repeating the
 *         request, with the same data and sequence number, up to three times.
 *
 * @param  ack_seq_nr   The sequence number of the TCP RX NOTIFY request we want to acknowledge.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_ip_tcp_rx_notify_ack(mp_seq_nr ack_seq_nr)
{
    P_WARN("[%02X]  -> Replying to TCP RX NOTIFY ...\n", ack_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_TCP_RX_NOTIFY, ack_seq_nr, 0x80, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Initiates a HTTP client connection to a remote HTTP server.
 *
 * @param  seq_nr       The message sequence number of the request sent to the BitPipe, if successful.
 * @param  secure       Specifies if the HTTP connection should be encrypted using SSL/TLS (i.e.: HTTPS). Values: [0,1]
 * @param  method       HTTP method to use. 0=GET, 1=POST, 2=HEAD.
 * @param  host_len     Length of the remote HTTP server hostname, or IP address (excluding NULL string terminator, if any).
 * @param  host         Remote HTTP server hostname, or IP address.
 * @param  port         HTTP remote port to connect to. If 0, 80 or 443 will be automatically chosen depending on the secure parameter.
 * @param  path_q_len   Length of the HTTP path and query parameters (excluding NULL string terminator, if any).
 * @param  path_q       The HTTP path and query parameters
 * @param  add_h_len    Length of additional headers to include in the HTTP request (excluding NULL string terminator, if any).
 * @param  add_h        Additional headers to include in the HTTP request. Please read documentation on how to properly format
 *                      the string when more than one header is specified.
 * @param  post_content_len Only used when method is POST. The length of the POST content that will be sent to the HTTP server.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_ip_http_init_connection(mp_seq_nr* seq_nr, uint8_t secure, uint8_t method, uint8_t host_len, uint8_t* host,
                                       uint16_t port, uint8_t path_q_len, uint8_t* path_q, uint8_t add_h_len, uint8_t* add_h, uint32_t post_content_len)
{
    uint8_t linear_buf[10+host_len+path_q_len+add_h_len];
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    linear_buf[0] = 0;
    if(secure)
    {
        linear_buf[0] |= (1<<0);        // SSL/TLS
    }
    linear_buf[0] |= (method << 1);

    linear_buf[1] = host_len;
    memcpy(&linear_buf[2], host, host_len);

    linear_buf[2+host_len] = (port >> 8) & 0xFF;
    linear_buf[3+host_len] = port & 0xFF;

    linear_buf[4+host_len] = path_q_len;
    memcpy(&linear_buf[5+host_len], path_q, path_q_len);

    linear_buf[5+host_len+path_q_len] = add_h_len;
    memcpy(&linear_buf[6+host_len+path_q_len], add_h, add_h_len);

    linear_buf[6+host_len+path_q_len+add_h_len] = (post_content_len >> 24) & 0xFF;
    linear_buf[7+host_len+path_q_len+add_h_len] = (post_content_len >> 16) & 0xFF;
    linear_buf[8+host_len+path_q_len+add_h_len] = (post_content_len >> 8) & 0xFF;
    linear_buf[9+host_len+path_q_len+add_h_len] = post_content_len & 0xFF;

    P_WARN("[%02X] Initiating %sHTTP connection to %s:%d ...\n", msg_seq_nr, secure ? "secure ":"", host, port);

    uint8_t status = bp_api_send_message(MSG_ID_HTTP_INITIATE_CONNECTION, msg_seq_nr, 0x00, linear_buf, sizeof(linear_buf));
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Appends and/or sends HTTP POST data to the HTTP server endpoint.
 *         This is only used when a HTTP request with method POST was initiated.
 *         Depending on the value of the 'write_op' parameter, this request does one of the following:
 * - write_op=0: Appends to BitPipe internal buffer the POST data. Can be used several times.
 * - write_op=1: Same as 0, but also sends the buffer on the Internet to the HTTP server. Can be used seevral times.
 * - write_op=2: Same as 1, but marks the POST data as being complete. At this point, the number of bytes sent should match
 *               the 'post_content_len' parameter used in the HTTP INITIATE CONNECTION request.
 *
 * @param  seq_nr       The message sequence number of the request sent to the BitPipe, if successful.
 * @param  http_conn_id A valid HTTP connection identifier acquired in the reply to the HTTP INITIATE CONNECTION request.
 * @param  write_op     The write operation to do in this request. Values: [0-2]
 * @param  data_len     Length of the user POST data in this message.
 * @param  data         User POST data.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_ip_http_post_write(mp_seq_nr* seq_nr, uint8_t http_conn_id, uint8_t write_op, uint8_t data_len, uint8_t* data)
{
    uint8_t linear_buf[3+data_len];
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    linear_buf[0] = http_conn_id;
    linear_buf[1] = write_op;
    linear_buf[2] = data_len;
    memcpy(&linear_buf[3], data, data_len);

    P_WARN("[%02X] Writing POST data to HTTP connection ...\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_HTTP_POST_WRITE, msg_seq_nr, 0x00, linear_buf, sizeof(linear_buf));
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Sends a reply (as acknowledgement) to the BitPipe, indicating that the HTTP RX NOTIFY request was well
 *         received by the Host.
 *
 * @note   Failure to acknowledge HTTP RX NOTIFY requests from the BitPipe will result in the BitPipe repeating the
 *         request, with the same data and sequence number, up to three times.
 *
 * @param  ack_seq_nr   The sequence number of the HTTP RX NOTIFY request we want to acknowledge.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_ip_http_rx_notify_ack(mp_seq_nr ack_seq_nr)
{
    P_WARN("[%02X]  -> Replying to HTTP RX NOTIFY ...\n", ack_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_HTTP_RX_NOTIFY, ack_seq_nr, 0x80, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Gracefully closes the TCP or HTTP connection identified by 'ip_conn_id'.
 *
 * @note   Never specify a MQTT connection identifier in this function. Use bp_api_mqtt_disconnect() instead.
 *
 * @param  seq_nr       The message sequence number of the request sent to the BitPipe, if successful.
 * @param  ip_conn_id   The TCP or HTTP connection identifier.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_ip_connection_close(mp_seq_nr* seq_nr, uint8_t ip_conn_id)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Closing IP connection (id=%d) ...\n", msg_seq_nr, ip_conn_id);

    uint8_t status = bp_api_send_message(MSG_ID_INET_CLOSE, msg_seq_nr, 0x00, &ip_conn_id, 1);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}


// ---------------------------------------------------------------------------------------------------------------------
// Doxygen documentation support.
// ---------------------------------------------------------------------------------------------------------------------


/**
 * @typedef void (*bp_api_ip_tcp_init_connection_reply)(uint8_t seq_nr, uint8_t status, uint8_t tcp_conn_id)
 * @brief  Callback into application when the reply to TCP INITIATE CONNECTION request is received from the BitPipe.
 *
 * @note   You will need to keep tack of the 'tcp_conn_id' for subsequent operations.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  tcp_conn_id  The connection identifier for this new connection.
 */

/**
 * @typedef void (*bp_api_ip_tcp_connection_established_notify_post)(uint8_t seq_nr, uint8_t tcp_conn_id)
 * @brief  Callback into application when the TCP CONNECTION ESTABLISHED NOTIFY message is posted from the BitPipe.
 *
 * @note   After connection establishment, you may now issue the bp_api_ip_tcp_write_send() function.
 *
 * @param  seq_nr       The sequence number of the message.
 * @param  tcp_conn_id  The connection whose status changed to 'established'.
 */

/**
 * @typedef void (*bp_api_ip_tcp_write_append_reply)(uint8_t seq_nr, uint8_t status, uint16_t buf_sz_current, uint16_t buf_sz_max)
 * @brief  Callback into application when the reply to TCP WRITE APPEND request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  buf_sz_current The current number of bytes used in the internal TCP buffer.
 * @param  buf_sz_max   The size of the internal buffer in the BitPipe for this connection. This is constant.
 */

/**
 * @typedef void (*bp_api_ip_tcp_write_send_reply)(uint8_t seq_nr, uint8_t status)
 * @brief  Callback into application when the reply to TCP WRITE SEND request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @typedef void (*bp_api_ip_tcp_rx_notify_request)(uint8_t seq_nr, uint8_t tcp_conn_id, uint8_t data_len, uint8_t* data)
 * @brief  Callback into application when the TCP RX NOTIFY message is requested from the BitPipe.
 *
 * This message is used whenever TCP data is received from the remote TCP server by the BitPipe.
 * You **MUST** reply to this message.
 *
 * @param  seq_nr       The sequence number of the message, which must be used in the reply to this request.
 * @param  tcp_conn_id  The TCP connection identifier, from where the data came from.
 * @param  data_len     Length of TCP data this message contains.
 * @param  data         The TCP data received from the TCP server.
 */

/**
 * @typedef void (*bp_api_ip_http_init_connection_reply)(uint8_t seq_nr, uint8_t status, uint8_t http_conn_id)
 * @brief  Callback into application when the reply to HTTP INITIATE CONNECTION request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  http_conn_id The connection identifier for this new HTTP request.
 */

/**
 * @typedef void (*bp_api_ip_http_post_write_ready_notify_post)(uint8_t seq_nr, uint8_t http_conn_id)
 * @brief  Callback into application when the HTTP POST WRITE READY NOTIFY message is posted from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message.
 * @param  http_conn_id The connection whose status now allows the POST data to be written.
 */

/**
 * @typedef void (*bp_api_ip_http_post_write_reply)(uint8_t seq_nr, uint8_t status)
 * @brief  Callback into application when the reply to HTTP POST WRITE request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @typedef void (*bp_api_ip_http_rx_notify_request)(uint8_t seq_nr, uint8_t http_conn_id, uint8_t data_len, uint8_t* data)
 * @brief  Callback into application when the HTTP RX NOTIFY message is requested from the BitPipe.
 *
 * This message is used whenever HTTP data is received from the remote HTTP server by the BitPipe.
 * You **MUST** reply to this message.
 *
 * @param  seq_nr       The sequence number of the message, which must be used in the reply to this request.
 * @param  http_conn_id The HTTP connection identifier, from where the data came from.
 * @param  data_len     Length of HTTP data this message contains.
 * @param  data         The HTTP data received from the HTTP server.
 */

/**
 * @typedef void (*bp_api_ip_connection_close_reply)(uint8_t seq_nr, uint8_t status)
 * @brief  Callback into application when the reply to INET CONNECTION CLOSE request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @typedef void (*bp_api_ip_connection_close_notify_post)(uint8_t seq_nr, uint8_t ip_conn_id, uint8_t reason)
 * @brief  Callback into application when the INET CONNECTION CLOSE NOTIFY message is posted from the BitPipe.
 *
 * You may not use the connection identifier 'ip_conn_id' in any subsequent operation.
 *
 * @param  seq_nr       The sequence number of the message.
 * @param  ip_conn_id   The connection identifier that is now closed.
 * @param  reason       The reason why the connection was closed. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @}       // Doxygen api_ip defgroup
 */

#ifdef __cplusplus
}
#endif
