/** ********************************************************************************************************************
 * @file        bp_api_mobi.c
 * @author      Briowireless, Inc.
 * @date        2016-11-02
 * @brief       API layer for MOBILE INFORMATION requests to the BitPipe.
 *              See the BitPipe Serial API document for more information on requests below.
 * @version     API V1
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/

// Include directives  -------------------------------------------------------------------------------------------------

#include <bp_api_mobi.h>
#include <bp_api.h>
#include <message_defs.h>

#ifdef __cplusplus
extern "C" {
#endif

// Preprocessor definitions --------------------------------------------------------------------------------------------

// Global (extern) function  -------------------------------------------------------------------------------------------

/**
 * @defgroup api_mobi BitPipe API: Mobile Information
 * Allows interacting with the wireless mobile technology on the BitPipe. Handles powering-up the device and querying
 *  information about the state of the device, and the state of the mobile network.
 * @{
 */

/** ********************************************************************************************************************
 * @brief  Powers-on the wireless device (Radio) on the BitPipe.
 *         If friendly network check ('back_off_enabled' parameter) is enabled, the following rules are enforced:
 *         - You may not do a power-on radio operation more than 6 times per hour; and
 *         - You must wait at least 5 minutes before each successful power-on radio request.
 *
 * @param  seq_nr           The message sequence number of the request sent to the BitPipe, if successful.
 * @param  pwr_on_method    Specifies the way to power-on the wireless device via VIN:
 *          - @ref BP_API_PWRON_RADIO__PWRON_METHOD_DCDC
 *          - @ref BP_API_PWRON_RADIO__PWRON_METHOD_BYPASS
 * @param  back_off_enabled Specifies if friendly network check is done (may prevent power-on). Values: [0,1]
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_mobi_power_on_radio(mp_seq_nr* seq_nr, uint8_t pwr_on_method, uint8_t back_off_enabled)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;
    const uint8_t byte4 = (uint8_t)(pwr_on_method & 0x7F) | (uint8_t)((back_off_enabled) << 7);

    P_WARN("[%02X] Powering On Radio %s..\n", msg_seq_nr, back_off_enabled ? "with back-off check " : "(BACKOFF CHECK DISABLED)");

    uint8_t status = bp_api_send_message(MSG_ID_POWER_ON_RADIO, msg_seq_nr, 0x00, &byte4, 1);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Powers-off the wireless device (Radio) on the BitPipe.
 * @note   You should always, if possible, use the graceful power-off radio method.
 *
 * @param  seq_nr           The message sequence number of the request sent to the BitPipe, if successful.
 * @param  pwr_off_method   Specifies the way to power-off the wireless device:
 *          - @ref BP_API_PWROFF_RADIO__PWROFF_METHOD_GRACEFUL (Recommended) The radio disconnects from the mobile
 *                 network, and shuts-down cleanly.
 *          - @ref BP_API_PWROFF_RADIO__PWROFF_METHOD_HARD The radio power is immediately cut. This method should be
 *                 avoided, as it does not inform mobile network of disconnection, which is not considered good practice.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_mobi_power_off_radio(mp_seq_nr* seq_nr, uint8_t pwr_off_method)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Powering Off Radio ..\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_POWER_OFF_RADIO, msg_seq_nr, 0x00, &pwr_off_method, 1);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Retrieves the current BitPipe status regarding the radio state.
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_mobi_get_radio_status(mp_seq_nr* seq_nr)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Querying Radio Status ..\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_GET_RADIO_STATUS, msg_seq_nr, 0x00, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Retrieves the International Mobile Equipment Identifier (IMEI) of the BitPipe.
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_mobi_get_imei(mp_seq_nr* seq_nr)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Querying IMEI ..\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_GET_IMEI, msg_seq_nr, 0x00, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Retrieves the Integrated Circuit Card Identifier (ICCID) of the SIM card in the BitPipe.
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_mobi_get_iccid(mp_seq_nr* seq_nr)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Querying ICCID ..\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_GET_ICCID, msg_seq_nr, 0x00, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Retrieves the International Mobile Subscriber Identifier (IMSI) of the SIM card in the BitPipe.
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_mobi_get_imsi(mp_seq_nr* seq_nr)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Querying IMSI ..\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_GET_IMSI, msg_seq_nr, 0x00, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Retrieves the mobile phone number associated to the SIM card in the BitPipe.
 * @note   This information needs to be provisionned in the SIM card for this operation to work.
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_mobi_get_own_phone_number(mp_seq_nr* seq_nr)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Querying Own Phone Number ..\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_GET_OWN_PHONE_NUMBER, msg_seq_nr, 0x00, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Retrieves the APN information used by the BitPipe to establish data links with the mobile operator.
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_mobi_get_apn_config(mp_seq_nr* seq_nr)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Querying Configured APN  ..\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_GET_APN_CONFIG, msg_seq_nr, 0x00, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Sets the APN information to use to establish data links with the mobile operator.
 * @note   This information is kept in non-volatile memory and loaded at BitPipe boot.
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 * @param  host_len Length of the APN name (excluding NULL string terminator, if any).
 * @param  host     The APN name.
 * @param  user_len The length of the APN user name (excluding NULL string terminator, if any).
 * @param  user     APN user name.
 * @param  pass_len The length of the APN password (excluding NULL string terminator, if any).
 * @param  pass     APN password.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_mobi_set_apn_config(mp_seq_nr* seq_nr, uint8_t host_len, uint8_t* host, uint8_t user_len, uint8_t* user, uint8_t pass_len, uint8_t* pass)
{
    uint8_t linear_buf[3+host_len+user_len+pass_len];
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    linear_buf[0] = host_len;
    memcpy(&linear_buf[1], host, host_len);
    linear_buf[1+host_len] = user_len;
    memcpy(&linear_buf[2+host_len], user, user_len);
    linear_buf[2+host_len+user_len] = pass_len;
    memcpy(&linear_buf[3+host_len+user_len], pass, pass_len);

    P_WARN("[%02X] Configure APN  ..\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_SET_APN_CONFIG, msg_seq_nr, 0x00, linear_buf, sizeof(linear_buf));
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Allows providing SIM card PIN (and/or PUK).
 *         This is required if the SIM card in the BitPipe is locked by a PIN.
 *         After 3 consecutive bad PINs, the SIM card may lock-up. After this, you must provide the PIN unlock key (PUK).
 * @see the function bp_api_mobi_get_radio_status_reply parameter 'sim_status'
 * @see the function bp_api_mobi_radio_start_notify_post parameter 'sim_status'.
 *
 * @note   This information is kept in non-volatile memory and loaded at BitPipe boot.
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 * @param  pin_len  Length of the PIN or PUK (excluding NULL string terminator, if any).
 * @param  pin      The PIN or PUK.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_mobi_set_sim_access_config(mp_seq_nr* seq_nr, uint8_t pin_len, uint8_t* pin)
{
    uint8_t linear_buf[1+pin_len];
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    linear_buf[0] = pin_len;
    memcpy(&linear_buf[1], pin, pin_len);

    P_WARN("[%02X] Setting SIM PIN ACCESS ..\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_SET_SIM_ACCESS_CONFIG, msg_seq_nr, 0x00, linear_buf, sizeof(linear_buf));
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Retrieves the current mobile connection status, serving cell IDs, link technology, signal strength.
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_mobi_get_mobile_conn_status(mp_seq_nr* seq_nr)
{

    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Querying Mobile Connection Status ..\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_GET_MOBILE_CONN_STATUS, msg_seq_nr, 0x00, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Retrieves the current mobile date and time.
 *
 * @param  seq_nr   The message sequence number of the request sent to the BitPipe, if successful.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_mobi_get_mobile_date_time(mp_seq_nr* seq_nr)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Querying Mobile Date and Time ..\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_GET_MOBILE_DATE_AND_TIME, msg_seq_nr, 0x00, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

// ---------------------------------------------------------------------------------------------------------------------
// Doxygen documentation support.
// ---------------------------------------------------------------------------------------------------------------------


/**
 * @typedef void (*bp_api_mobi_power_on_radio_reply)(uint8_t seq_nr, uint8_t status)
 * @brief  Callback into application when the reply to POWER ON RADIO request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @typedef void (*bp_api_mobi_power_off_radio_reply)(uint8_t seq_nr, uint8_t status)
 * @brief  Callback into application when the reply to POWER OFF RADIO request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @typedef void (*bp_api_mobi_radio_start_notify_post)(uint8_t seq_nr, uint8_t notify_reason, uint8_t sim_status)
 * @brief  Callback into application when the RADIO START NOTIFY message is posted from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message.
 * @param  notify_reason    The reason for this posted notification. See the *BitPipe Serial API* documentation for possible values.
 * @param  sim_status   The status of the SIM card.
 */

/**
 * @typedef void (*bp_api_mobi_get_radio_status_reply)(uint8_t seq_nr, uint8_t status, uint8_t radio_status, uint8_t sim_status)
 * @brief  Callback into application when the reply to GET RADIO STATUS request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  radio_status The status of the radio. See the *BitPipe Serial API* documentation for possible values.
 * @param  sim_status   The status of the SIM card. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @typedef void (*bp_api_mobi_get_imei_reply)(uint8_t seq_nr, uint8_t status, uint8_t* imei)
 * @brief  Callback into application when the reply to GET IMEI request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  imei         The IMEI of the BitPipe (15 ASCII characters).
 */

/**
 * @typedef void (*bp_api_mobi_get_iccid_reply)(uint8_t seq_nr, uint8_t status, uint8_t* iccid)
 * @brief  Callback into application when the reply to GET ICCID request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  iccid        The ICCID of the SIM card in the BitPipe (20 ASCII characters).
 */

/**
 * @typedef void (*bp_api_mobi_get_imsi_reply)(uint8_t seq_nr, uint8_t status, uint8_t* imsi)
 * @brief  Callback into application when the reply to GET IMSI request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  imsi         The IMSI of the SIM card in the BitPipe (15 ASCII characters).
 */

/**
 * @typedef void (*bp_api_mobi_get_own_phone_number_reply)(uint8_t seq_nr, uint8_t status, uint8_t msisdn_len, uint8_t* msisdn)
 * @brief  Callback into application when the reply to GET OWN PHONE NUMBER request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  msisdn_len   The length of the phone number.
 * @param  msisdn       The phone number.
 */

/**
 * @typedef void (*bp_api_mobi_get_apn_config_reply)(uint8_t seq_nr, uint8_t status, uint8_t host_len, uint8_t* host, uint8_t user_len, uint8_t* user, uint8_t pass_len, uint8_t* pass)
 * @brief  Callback into application when the reply to GET APN CONFIGURATION request is received from the BitPipe.
 *
 * @param  seq_nr   The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status   The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  host_len Length of the APN name (excluding NULL string terminator, if any).
 * @param  host     The APN name.
 * @param  user_len The length of the APN user name (excluding NULL string terminator, if any).
 * @param  user     APN user name.
 * @param  pass_len The length of the APN password (excluding NULL string terminator, if any).
 * @param  pass     APN password.
 */

/**
 * @typedef void (*bp_api_mobi_set_apn_config_reply)(uint8_t seq_nr, uint8_t status)
 * @brief  Callback into application when the reply to SET APN CONFIGURATION request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @typedef void (*bp_api_mobi_set_sim_access_config_reply)(uint8_t seq_nr, uint8_t status)
 * @brief  Callback into application when the reply to SET SIM ACCESS request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @typedef void (*bp_api_mobi_get_mobile_conn_status_reply)(uint8_t seq_nr, uint8_t status, uint8_t conn_ind, uint8_t link_tech, uint16_t net_lac, uint32_t net_cell_id, uint8_t signal_strength, uint8_t* net_mcc, uint8_t* net_mnc)
 * @brief  Callback into application when the reply to GET MOBILE CONNECTION STATUS request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  conn_ind     Connected indicator:
 * - 0: Not connected.
 * - 1: Camping on a mobile cell; not registered.
 * - 2: Registered on home mobile network.
 * - 3: Registered on foreign mobile network.
 * @param  link_tech    Link technology:
 * - 0: Unknown
 * - 1: 2G
 * - 2: 3G
 * - 3: LTE (4G)
 * @param  net_lac      Network location area code (16-bit value).
 * @param  net_cell_id  Network cell identifier (32-bit value).
 * @param  signal_strength  Signal strength:
 * - 255: Unknown
 * - 2G: 0-63, Subtract 110 from this value to have BCCH value in dBm.
 * - 3G: 0-99, Subtract 124 from this value to have RSCP value in dBm.
 * - 4G: 0-96, Subtract 140 from this value to have RSRP value in dBm.
 * @param  net_mcc      Mobile country code (4 ASCII characters).
 * @param  net_mnc      Mobile network code (4 ASCII characters).
 */

/**
 * @typedef void (*bp_api_mobi_mobile_conn_status_change_notify_post)(uint8_t seq_nr, uint8_t conn_ind, uint8_t link_tech, uint16_t net_lac, uint16_t net_cell_id)
 * @brief  Callback into application when the RADIO START NOTIFY message is posted from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message.
 * @param  conn_ind     Connected indicator:
 * - 0: Not connected.
 * - 1: Camping on a mobile cell; not registered.
 * - 2: Registered on home mobile network.
 * - 3: Registered on foreign mobile network.
 * @param  link_tech    Link technology:
 * - 0: Unknown
 * - 1: 2G
 * - 2: 3G
 * - 3: LTE (4G)
 * @param  net_lac      Network location area code (16-bit value).
 * @param  net_cell_id  Network cell identifier (32-bit value).
 */

/**
 * @typedef void (*bp_api_mobi_get_mobile_date_time_reply)(uint8_t seq_nr, uint8_t status, uint8_t year, uint8_t month, uint8_t day, uint8_t hour, uint8_t minute, uint8_t second)
 * @brief  Callback into application when the reply to GET MOBILE DATE AND TIME request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  year         Current date year, minus 2000.
 * @param  month        Current date month, 0-based.
 * @param  day          Current date day, 0-based.
 * @param  hour         Current time, in 24-hour format.
 * @param  minute       Current time minute.
 * @param  second       Current time second.
 */

/**
 * @}       // Doxygen api_mobi defgroup
 */

#ifdef __cplusplus
}
#endif
