/** ********************************************************************************************************************
 * @file        bp_api_sys.c
 * @author      Briowireless, Inc.
 * @date        2016-10-28
 * @brief       API layer for SYSTEM requests to the BitPipe.
 *              See the BitPipe Serial API document for more information on requests below.
 * @version     API V1
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/

// Include directives  -------------------------------------------------------------------------------------------------

#include <bp_api_sys.h>
#include <bp_api.h>
#include <message_defs.h>

#ifdef __cplusplus
extern "C" {
#endif

// Preprocessor definitions --------------------------------------------------------------------------------------------

// Global (extern) function  -------------------------------------------------------------------------------------------

/**
 * @defgroup api_sys BitPipe API: System
 * Allows configuration of the binary communication channel with the BitPipe.
 * Also allows receiving error reports from the BitPipe.
 * @{
 */

/** ********************************************************************************************************************
 * @brief Configures the maximum variable payload size the BitPipe can send to this application.
 *        This parameter is used by the BitPipe to control the size of messages sent to the Host for the following
 *        operations:
 *        - TCP Rx Notify
 *        - Http Rx Notify
 *        - MQTT Publication Rx Notify
 *        - SMS Message Part Notify
 *
 * @param  seq_nr           The message sequence number of the request sent to the BitPipe, if successful.
 * @param  max_payload_sz   Maximum message payload size. Values: [48-260]
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_sys_set_max_payload_sz(mp_seq_nr* seq_nr, uint16_t max_payload_sz)
{
    uint8_t data[] = { (uint8_t)((max_payload_sz >> 8) & 0xFF), (uint8_t)(max_payload_sz & 0xFF) };
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_INFO("[%02X] Configuring Max Message size (%d bytes)...\n", msg_seq_nr, max_payload_sz);

    uint8_t status = bp_api_send_message(MSG_ID_SET_MAX_MSG_PAYSIZE, msg_seq_nr, 0x00, data, sizeof(data));
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief Configures the maximum variable payload size the BitPipe can send to this application.
 *
 * @param  seq_nr       The message sequence number of the request sent to the BitPipe, if successful.
 * @param  uart_speed   The uart speed identifier (see documentation for more info).
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_sys_set_uart_speed(mp_seq_nr* seq_nr, uint8_t uart_speed)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_INFO("[%02X] Configuring UART Speed ...\n", msg_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_SET_UART_SPEED, msg_seq_nr, 0x00, &uart_speed, sizeof(uart_speed));
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}


// ---------------------------------------------------------------------------------------------------------------------
// Doxygen documentation support.
// ---------------------------------------------------------------------------------------------------------------------

/**
 * @typedef void (*bp_api_sys_error_report_post)(uint8_t seq_nr, uint8_t facility, uint8_t err_nr)
 * @brief  Callback into application when the ERROR REPORT message is posted from the BitPipe.
 *
 * @see the *Error Numbers and Meaning* section of the **BitPipe Serial API** documentation.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  facility     The error facility.
 * @param  err_nr       The specific error number, within the specified facility.
 */

/**
 * @typedef void (*bp_api_sys_set_max_payload_sz_reply)(uint8_t seq_nr, uint8_t status)
 * @brief  Callback into application when the reply to SET MAXIMUM MESSAGE PAYLOAD SIZE request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @typedef void (*bp_api_sys_set_uart_speed_reply)(uint8_t seq_nr, uint8_t status)
 * @brief  Callback into application when the reply to SET UART SPEED request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @}       // Doxygen api_sys defgroup
 */

#ifdef __cplusplus
}
#endif
