/** ********************************************************************************************************************
 * @file        message_protocol.h
 * @author      Brioconcept Consulting, Inc.
 * @date        2015-08-05
 * @brief       Protocol to send and receive messages from point to point.
 *              This implementation allows communication with multiple endpoints;
 *              A context is used for API calls.
 * @copyright   Brioconcept Consulting, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/
#ifndef _MESSAGE_PROTOCOL_H
#define _MESSAGE_PROTOCOL_H

#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

// Preprocessor definitions --------------------------------------------------------------------------------------------

#define MESSAGE_PROTOCOL_HEADER_SIZE        4                               ///< Header size: SOF, CMD_ID, STATUS, SEQ_NR
#define MESSAGE_PROTOCOL_FOOTER_SIZE        3                               ///< Footer size: EOF, CRC1, CRC2


// Global Type Defintions ---------------------------------------------------------------------------------------------

/// @brief Internal parser state
///        This enum is only provided to obtain its size (sizeof) when defining buffer size in communication layer.
typedef enum __mp_parser_state
{
    STATE_IDLE = 0,
    STATE_IDLE_ESCAPE_DATA,
    STATE_WAIT_DATA,
    STATE_ESCAPE_DATA,
    STATE_MESSAGE_READY
} mp_parser_state;

/// @brief This structure holds information about messaging.
///        It contains parser state and work buffer information.
typedef struct __mp_context
{
    uint8_t* work_buf;
    uint16_t work_buf_len;
    uint16_t work_buf_idx;
    uint16_t work_buf_crc;
    mp_parser_state parser_state;
} mp_context;

/// @brief Type for sequence numbers used in the message_protocol API.
typedef uint8_t mp_seq_nr;

/// @brief Definition of a message.
typedef struct __mp_message
{
    uint8_t  cmd_id;                        ///< Command identifier.
    uint8_t  seq_nr;                        ///< Message sequence number.
    uint8_t  status;                        ///< Status.
    uint8_t* data;                          ///< Message data.
    uint16_t data_len;                      ///< Length of message data.
} mp_message;

/// @brief message_protocol API return values.
typedef enum __mp_status
{
    MESSAGE_PROTOCOL_OK              = 0,   ///< Operation was successful.
    MESSAGE_PROTOCOL_MESSAGE_READY,         ///< A message is ready. See message_protocol_push_data() for its usage.

    MESSAGE_PROTOCOL_WARN_NOT_AVAILABLE,    ///< Warning: No message is ready to be retrieved. See message_protocol_retrieve().
    MESSAGE_PROTOCOL_ERROR_ARGS,            ///< Warning: Invalid arguments were specified.
    MESSAGE_PROTOCOL_ERROR_BUF_TOO_SMALL,   ///< Warning: The buffer provided is too small.
    MESSAGE_PROTOCOL_ERROR_INDEX,           ///< Warning: The message is invalid. Should not occur.
    MESSAGE_PROTOCOL_ERROR_CHECKSUM,        ///< Warning: Indicates a message is ready but its checksum is invalid. The
                                            ///            message bytes are discarded, and the message parser is reset.
    MESSAGE_PROTOCOL_ERROR_UNKOWN_STATE     ///< Error: Indicates an unexpected internal state. A bug is at work here. This error is not recoverable.
} mp_status;



// Global Functions Prototypes ----------------------------------------------------------------------------------------

// Takes ownership of the buffer specified, creates a new context, and initializes the message protocol internals.
// - The 'buffer' parameter must be sufficiently large to hold the context data, and the largest message you expect to
//   receive, including header, footer, escape characters and data.
mp_status   message_protocol_init(mp_context *ctx, uint8_t* buffer, uint16_t buffer_len);

// Pushes one byte of received message data into the internal message parser.
mp_status   message_protocol_push_data(mp_context *ctx, uint8_t message_byte);

// Retrieves the message received (after 'message_protocol_push_data' returns 'MESSAGE_PROTOCOL_MESSAGE_READY'.
// IMPORTANT: The received message data is _NOT_ copied into this message; i.e.: the 'data' member points to the data
//            in the work buffer.
mp_status   message_protocol_retrieve(mp_context *ctx, mp_message* const message);

// Serializes the specified message in the buffer provided.
// - The buffer_len is used to prevent buffer overflow.
mp_status   message_protocol_serialize(const mp_message* message, uint8_t* buffer, uint16_t buffer_len, uint16_t* buffer_used_len);

// Use this function to deep copy a message. No allocation is made.
// - The 'clone_data_len' parameter is explicitely redundant and describes the length of the buffer pointed to by
//   clone->data. This buffer must be at least 'from->data_len' long.
mp_status   message_protocol_clone(const mp_message* from, mp_message* const clone, uint16_t clone_data_len);


#ifdef __cplusplus
}
#endif


#endif  // _MESSAGE_PROTOCOL_H
